from realnet.core.provider import OrgsProvider
from realnet.core.type import Org, Account, Authenticator, Client
from .models import Org as OrgModel, Account as AccountModel, Authenticator as AuthenticatorModel, Client as ClientModel, Item as ItemModel, Type as TypeModel, Acl,  AclType, session as db
from .utility import item_model_to_item, get_types_by_name, type_model_to_type, get_derived_types
from sqlalchemy import or_

class SqlOrgsProvider(OrgsProvider):

    def get_orgs(self):
            return [Org(org.id, org.name) for org in db.query(OrgModel).all()]

    def get_org_by_id(self,id):
        org = db.query(OrgModel).filter(OrgModel.id == id).first()
        if org:
            return Org(org.id, org.name)
        return None

    def get_org_by_name(self,name):
        org = db.query(OrgModel).filter(OrgModel.name == name).first()
        if org:
            return Org(org.id, org.name)
        return None

    def get_account_by_id(self,id):
        account = db.query(AccountModel).get(id)
        if account:
            return Account(account.id, account.username, Org(account.org.id, account.org.name))
        else:
            return None

    def check_password(self, org_id, account_id, password):
        account = db.query(AccountModel).filter(AccountModel.org_id == org_id, AccountModel.username == account_id).first()
        if account:
            if password and account.check_password(password):
                return Account(account.id, account.username, Org(account.org.id, account.org.name))
        return None

    def get_org_authenticators(self, org_id):
        org = db.query(OrgModel).filter(OrgModel.name == org_id).first()
        if not org:
            # is there a public org?
            org = db.query(OrgModel).filter(OrgModel.public == True).first()
        if org:
            return [Authenticator(a.name,a.get_url()) for a in db.query(AuthenticatorModel).filter(AuthenticatorModel.org_id == org.id).all()]
        return []

    def get_org_clients(self, org_id):
        return [Client( client.id, 
                        client.name, 
                        Org(client.org.id, client.org.name), 
                        client.attributes) for client in db.query(ClientModel).filter(ClientModel.org_id == org_id).all()]

    def get_org_client(self, org_id, client_id):
        return db.query(ClientModel).filter(or_(ClientModel.client_id == client_id, ClientModel.name == client_id),ClientModel.org_id == org_id).first()

    def get_public_apps(self, org_id):
        tbn = get_types_by_name(org_id)
        type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name == 'App', TypeModel.org_id == org_id).all()]
        derived_type_ids = get_derived_types(org_id, type_ids)
        apps = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))), ItemModel.org_id == org_id).all()
        return [item_model_to_item(org_id, app, tbn) for app in apps]

    def get_public_types(self, org_id):
        tbn = get_types_by_name(org_id)
        return tbn.values()

    def get_public_forms(self, org_id):
        tbn = get_types_by_name(org_id)
        type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name == 'Form', TypeModel.org_id == org_id).all()]
        derived_type_ids = get_derived_types(org_id, type_ids)
        forms = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))), ItemModel.org_id == org_id).all()
        return [item_model_to_item(org_id, form, tbn) for form in forms]

    def get_public_orgs(self):
        return [Org(org.id, org.name) for org in db.query(OrgModel).filter(OrgModel.public == True).all()]

    def get_public_item(self, id):
        item = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.id == id).first()
        if item:
            tbn = get_types_by_name(item.org_id)
            return item_model_to_item(item.org_id, item, tbn)
        return None


import uuid
from realnet.core.type import Type
from realnet.provider.sql.models import Type as TypeModel, session as db
from realnet.core.provider import TypeProvider
from .utility import get_types_by_name, create_instance_model, create_type_model, instance_model_to_instance, get_derived_types

class SqlTypeProvider(TypeProvider):

    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_types(self):
        return [t for t in get_types_by_name(self.org_id).values()]

    def get_derived_types(self, type_ids):
        return get_derived_types(self.org_id, type_ids)

    def get_type_by_id(self, id):
        return {t.id:t for t in get_types_by_name(self.org_id).values()}.get(id)
        
    def get_type_by_name(self, name):
        return get_types_by_name(self.org_id).get(name)

    def get_type_instances(self, id):
        pass

    def delete_type(self, id):
        pass

    def update_type(self, id, **kwargs):
        pass

    def create_type(self, **kwargs):
        type_id=str(uuid.uuid4())
        type_name = None
        type_icon = None
        type_attributes = None
        type_base_id = None
        type_module = None

        for key, value in kwargs.items():
            # print("%s == %s" % (key, value))
            if key == 'name':
                type_name = value
            elif key == 'id':
                type_id = value
            elif key == 'base_id':
                type_base_id = value
            elif key == 'base':
                type = db.query(TypeModel).filter(TypeModel.name == value, TypeModel.org_id == self.org_id).first()
                if type:
                    type_base_id = type.id
            elif key == 'attributes':
                type_attributes = value
            elif key == 'icon':
                type_icon = value
            elif key == 'module':
                type_module = value

        type = create_type_model(
            db=db,
            type_id=type_id,
            type_name=type_name,
            type_icon=type_icon,
            type_attributes=type_attributes,
            owner_id=self.account_id,
            org_id=self.org_id,
            type_module=type_module,
            type_base_id=type_base_id)

        db.add(type)
        db.commit()
        
        return self.get_type_by_id(type.id)

    def create_instance(self, **kwargs):
        instance_id=str(uuid.uuid4())
        instance_name = None
        instance_icon = None
        instance_public = False
        instance_attributes = None
        instance_type_id = None
        instance_parent_type_id = None

        types_by_name = get_types_by_name(self.org_id)
        types_by_id = {t.id:t for t in types_by_name.values()}

        for key, value in kwargs.items():
            # print("%s == %s" % (key, value))
            if key == 'name':
                instance_name = value
            elif key == 'id':
                instance_id = value
            elif key == 'type_id':
                type = types_by_id.get(value)
                if type:
                    instance_type_id = type.id
            elif key == 'type':
                type = types_by_name.get(value)
                if type:
                    instance_type_id = type.id
            elif key == 'parent_type_id':
                type = types_by_id.get(value)
                if type:
                    instance_parent_type_id = type.id
            elif key == 'parent_type':
                type = types_by_name.get(value)
                if type:
                    instance_parent_type_id = type.id
            elif key == 'attributes':
                instance_attributes = value
                if value and 'icon' in value:
                    instance_icon = value['icon']
            elif key == 'icon':
                instance_icon = value
            elif key == 'public':
                instance_public = str(value).lower() == 'true'

        instance = create_instance_model(
            instance_id=instance_id,
            instance_name=instance_name,
            instance_icon=instance_icon,
            instance_attributes=instance_attributes,
            instance_public=instance_public,
            owner_id=self.account_id,
            org_id=self.org_id,
            instance_type_id=instance_type_id,
            instance_parent_type_id=instance_parent_type_id)

        db.add(instance)
        db.commit()
        
        return instance_model_to_instance(instance, types_by_name)
        


from flask import render_template
from realnet.resource.items.items import Items

class Types(Items):
    
    def get_endpoint_name(self):
        return 'types'

    def get_items(self, module, account, query, parent_item=None):
        types = [t for t in module.get_types()]
        for t in types:
            t.attributes['resource'] = 'types'
        return types

    def get_item(self, module, account, path):
        item = module.get_type_by_id(path)
        if item:
            return item

        return None


from authlib.integrations.flask_oauth2 import (
    AuthorizationServer,
    ResourceProtector,
)
from authlib.integrations.sqla_oauth2 import (
    create_query_client_func,
    create_save_token_func,
    create_revocation_endpoint,
    create_bearer_token_validator,
)
from authlib.oauth2.rfc6749 import grants
from authlib.oauth2.rfc7636 import CodeChallenge
from authlib.oidc.core import UserInfo
import authlib.oidc.core.grants as oidc_grants
from werkzeug.security import gen_salt

from realnet.provider.sql.models import db, Account, Client, AuthorizationCode, Token, session 


class AuthorizationCodeGrant(grants.AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        'client_secret_basic',
        'client_secret_post',
        'none',
    ]

    def save_authorization_code(self, code, request):
        print('*** authorization_code_grant:save_authorization_code ***')
        code_challenge = request.data.get('code_challenge')
        code_challenge_method = request.data.get('code_challenge_method')

        if request.data.get('nonce'):
            auth_code = AuthorizationCode(
                code=code,
                client_id=request.client.client_id,
                redirect_uri=request.redirect_uri,
                scope=request.scope,
                account_id=request.user.id,
                code_challenge=code_challenge,
                code_challenge_method=code_challenge_method,
                nonce=request.data.get('nonce'))
        else:
            auth_code = AuthorizationCode(
                code=code,
                client_id=request.client.client_id,
                redirect_uri=request.redirect_uri,
                scope=request.scope,
                account_id=request.user.id,
                code_challenge=code_challenge,
                code_challenge_method=code_challenge_method)

        db.session.add(auth_code)
        db.session.commit()
        return auth_code

    def query_authorization_code(self, code, client):
        print('*** authorization_code_grant:query_authorization_code ***')
        auth_code = AuthorizationCode.query.filter_by(
            code=code, client_id=client.client_id).first()
        if auth_code and not auth_code.is_expired():
            return auth_code

    def delete_authorization_code(self, authorization_code):
        print('*** authorization_code_grant:delete_authorization_code ***')
        db.session.delete(authorization_code)
        db.session.commit()

    def authenticate_user(self, authorization_code):
        print('*** authorization_code_grant:authenticate_user ***')
        return Account.query.get(authorization_code.account_id)


class PasswordGrant(grants.ResourceOwnerPasswordCredentialsGrant):
    def authenticate_user(self, username, password):
        account = Account.query.filter_by(username=username).first()
        print(account.check_password(password))
        if account is not None and account.check_password(password):
            return account


class RefreshTokenGrant(grants.RefreshTokenGrant):
    def authenticate_refresh_token(self, refresh_token):
        token = Token.query.filter_by(refresh_token=refresh_token).first()
        if token and token.is_refresh_token_active():
            return token

    def authenticate_user(self, credential):
        return Account.query.get(credential.user_id)

    def revoke_old_credential(self, credential):
        credential.revoked = True
        db.session.add(credential)
        db.session.commit()

from realnet.core.config import Config

config = Config()

class OpenIDCode(oidc_grants.OpenIDCode):
    def exists_nonce(self, nonce, request):
        print('*** openid_code_grant:exists_nonce ***')
        exists = AuthorizationCode.query.filter_by(
            client_id=request.client_id, nonce=nonce
        ).first()
        return bool(exists)

    def get_jwt_config(self, grant):
        print('*** openid_code_grant:get_jwt_config ***')
        return {
            'key': config.get_jwt_key(),
            'alg': 'HS256',
            'iss': config.get_jwt_issuer(),
            'exp': 3600
        }

    def generate_user_info(self, user, scope):
        print('*** openid_code_grant:get_user_info ***')
        user_info = UserInfo(sub=user.id, name=user.username)
        if 'email' in scope:
            user_info['email'] = user.email
        return user_info

class OpenIDImplicitGrant(oidc_grants.OpenIDImplicitGrant):
    def exists_nonce(self, nonce, request):
        print('*** openid_implicit_grant:exists_nonce ***')
        exists = AuthorizationCode.query.filter_by(
            client_id=request.client_id, nonce=nonce
        ).first()
        return bool(exists)

    def get_jwt_config(self):
        print('*** openid_implicit_grant:get_jwt_config ***')
        return {
            'key': config.get_jwt_key(),
            'alg': 'HS256',
            'iss': config.get_jwt_issuer(),
            'exp': 3600
        }

    def generate_user_info(self, user, scope):
        print('*** openid_implicit_grant:generate_user_info ***')
        user_info = UserInfo(sub=user.id, name=user.name)
        if 'email' in scope:
            user_info['email'] = user.email
        return user_info

def create_authorization_code(client, grant_user, request):
    code = gen_salt(48)
    nonce = request.data.get('nonce')
    auth_code = AuthorizationCode(
        code=code,
        client_id=request.client.client_id,
        redirect_uri=request.redirect_uri,
        scope=request.scope,
        user_id=request.user.id,
        nonce=nonce)
    db.session.add(auth_code)
    db.session.commit()
    return code

class HybridGrant(oidc_grants.OpenIDHybridGrant):

    def create_authorization_code(self, client, grant_user, request):
        print('*** hybrid_grant:create_authorization_code ***')
        return create_authorization_code(client, grant_user, request)

    def exists_nonce(self, nonce, request):
        print('*** hybrid_grant:exists_nonce ***')
        exists = AuthorizationCode.query.filter_by(
            client_id=request.client_id, nonce=nonce
        ).first()
        return bool(exists)

    def get_jwt_config(self):
        print('*** hybrid_grant:get_jwt_config ***')
        return {
            'key': config.get_jwt_key(),
            'alg': 'RS512',
            'iss': config.get_jwt_issuer(),
            'exp': 3600
        }

    def generate_user_info(self, user, scope):
        print('*** hybrid_grant:generate_user_info ***')
        user_info = UserInfo(sub=user.id, name=user.name)
        if 'email' in scope:
            user_info['email'] = user.email
        return user_info


query_client = create_query_client_func(session, Client)
save_token = create_save_token_func(session, Token)
authorization = AuthorizationServer(
    query_client=query_client,
    save_token=save_token,
)
require_oauth = ResourceProtector()


def config_oauth(app):

    authorization.init_app(app)

    # support all grants
    authorization.register_grant(grants.ImplicitGrant)
    authorization.register_grant(OpenIDImplicitGrant)
    authorization.register_grant(grants.ClientCredentialsGrant)
    authorization.register_grant(AuthorizationCodeGrant, [CodeChallenge(required=True), OpenIDCode(require_nonce=True)])
    authorization.register_grant(HybridGrant)

    authorization.register_grant(PasswordGrant)
    authorization.register_grant(RefreshTokenGrant)

    # support revocation
    revocation_cls = create_revocation_endpoint(session, Token)
    authorization.register_endpoint(revocation_cls)

    # protect resource
    bearer_cls = create_bearer_token_validator(session, Token)
    require_oauth.register_token_validator(bearer_cls())


from flask import Response, redirect, render_template, render_template_string, request, jsonify, Blueprint, send_file, session, current_app, url_for
from authlib.integrations.flask_oauth2 import current_token
from authlib.integrations.flask_client import OAuth
from authlib.integrations.requests_client import OAuth2Session
from authlib.common.encoding import to_unicode, to_bytes
from .auth import require_oauth, authorization
from realnet.provider.sql.models import get_or_create_delegated_account

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from authlib.common.urls import url_encode

router_bp = Blueprint('router_bp',__name__)

def current_user(contextProvider):
    if 'id' in session:
        uid = session['id']
        return contextProvider.get_account_by_id(uid)
    return None

@router_bp.route('/signin', defaults={'org_name': None}, methods=['GET', 'POST'])
@router_bp.route('/<org_name>/signin', methods=['GET', 'POST'])
def login(org_name):
    contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
    if request.method == 'GET':
        org = contextProvider.get_org_by_name(org_name)
        if not org:
            # is there a public org?
            org = contextProvider.get_org_by_name('public')
        if org:
            authenticators = contextProvider.get_org_authenticators(org.id)
            return render_template('login.html', authenticators = authenticators)
        else:
            return render_template('login.html', error=True, msg='Login not permitted')
    else:
        org = contextProvider.get_org_by_name(org_name)
        if not org:
            # is there a public org?
            org = contextProvider.get_org_by_name('public')
        if org:
            # is there already an account with that username?
            username = request.form.get('username')
            if username:
                password = request.form.get('password')
                if password:
                    account = contextProvider.check_password(org.id, username, password)
                    if account:
                        session['id'] = account.id
                        return redirect('/')
                
                return render_template('login.html', error=True, msg='Invalid username or password')
            else:
                return render_template('login.html', error=True, msg='Invalid username or password')
        else:
            return render_template('login.html', error=True, msg='Login not permitted')

@router_bp.route('/<id>/login', defaults={'name': None}, methods=['GET', 'POST'] )
@router_bp.route('/<id>/login/<name>',methods=['GET', 'POST'] )
def tenant_login(id, name):
    # 1. get the org
    # print(request.url)
    contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
    org =  org = contextProvider.get_org_by_name(id)
    if org:
        client_id = request.args.get('client_id')
        response_type = request.args.get('response_type')
        client = contextProvider.get_org_client(org.id, client_id)
        if not client:
            client = [c for c in contextProvider.get_org_clients(org.id) if c.name.endswith("_realscape_web")][0]
            # client = [c for c in contextProvider.get_org_clients(org.id) if c.name.endswith("_cli")][0]
        if client:
            if request.method == 'POST':
                username = request.form.get('username')
                if username:
                    password = request.form.get('password')
                    if password:
                        account = contextProvider.check_password(org.id, username, password)
                        if account:
                            return authorization.create_authorization_response(grant_user=account)
            else:
                if name == None:
                    oauths = [{'name': n['name'],
                               'url': '/{0}/login/{1}?client_id={2}&response_type={3}'.format(id, n['name'], client_id, response_type)} for n in
                              [q.to_dict() for q in contextProvider.get_org_authenticators(org.id)]]

                    return render_template('login.html', authenticators=oauths, client_id=client_id)
                else:
                    auth  = next(filter(lambda auth: auth.name == name, contextProvider.get_org_authenticators(org.id)), None)
                    if auth:
                        print(auth)
                        oauth = OAuth(current_app)
                        data = auth.to_dict()
                        del data['name']
                        del data['id']
                        backend = oauth.register(auth.name, **data)
                        redirect_uri = url_for('router_bp.tenant_auth', _external=True, id=id, client_id=client_id, name=name, response_type=response_type)
                        print(redirect_uri)
                        return backend.authorize_redirect(redirect_uri=redirect_uri)
                    else:
                        return jsonify(isError=True,
                                       message="Failure",
                                       statusCode=404,
                                       data='Authenticator {0} not found'.format(name)), 404
        else:
            return jsonify(isError=True,
                           message="Failure",
                           statusCode=404,
                           data='Client {0} not found'.format(client_id)), 404

    return jsonify(isError=True,
                   message="Failure",
                   statusCode=404,
                   data='Tenant {0} not found'.format(id)), 404

@router_bp.route('/logout', methods=['POST'])
def logout():
    del session['id']
    return redirect('/')

def add_params_to_qs(query, params):
    """Extend a query with a list of two-tuples."""
    if isinstance(params, dict):
        params = params.items()

    qs = urlparse.parse_qsl(query, keep_blank_values=True)
    qs.extend(params)
    return url_encode(qs)

def add_params_to_uri(uri, params, fragment=False):
    """Add a list of two-tuples to the uri query components."""
    sch, net, path, par, query, fra = urlparse.urlparse(uri)
    if fragment:
        fra = add_params_to_qs(fra, params)
    else:
        query = add_params_to_qs(query, params)
    return urlparse.urlunparse((sch, net, path, par, query, fra))


@router_bp.route('/register', methods=('GET', 'POST'))
def tenant_register():
    if request.method == 'POST':
        username = request.form.get('username')
        email = request.form.get('email')
        password = request.form.get('password')
        #todo complete registration
        account1 = None #Account.query.filter_by(username=username).first()
        account2 = None #Account.query.filter_by( email=email).first()
        if account1 or account2:
            return render_template('register.html')
        else:
            # account = Account(id=str(uuid.uuid4()), username=username, email=email)
            # account.set_password(password)
            # db.session.add(account)
            # db.session.commit()
            # if user is not just to log in, but need to head back to the auth page, then go for it
            next_page = request.args.get('next')
            if next_page:
                return redirect(next_page)
            return redirect('/')
    else:
        return render_template('register.html')

@router_bp.route('/<id>/authorize/<name>', defaults={'client_id': None})
@router_bp.route('/<id>/<client_id>/authorize/<name>')
def tenant_auth(id, client_id, name):
    # 1. get the
    print(request.url)
    contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
    org =  org = contextProvider.get_org_by_name(id)
    if org:
        if client_id is None:
            client_id = request.args.get('client_id')
        client = contextProvider.get_org_client(org.id, client_id)
        if not client:
            client = [c for c in contextProvider.get_org_clients(org.id) if c.name.endswith("_realscape_web")][0]
        
        if client:
            auth = next(filter(lambda auth: auth.name == name, contextProvider.get_org_authenticators(org.id)), None)
            if auth:
                    data = auth.to_dict()
                    del data['name']
                    del data['id']
                    print(request)
                    code = request.args.get('code')
                    response_type = request.args.get('response_type')
                    if code:
                        oaclient = OAuth2Session(auth.client_id, auth.client_secret, scope=request.args.get('scope'))
                        token_endpoint = auth.access_token_url
                        try:
                            redirect_uri = url_for('orgs_bp.tenant_auth', _external=True, id=id, client_id=client_id, name=name, response_type=response_type)
                            token_test = oaclient.fetch_token(token_endpoint, authorization_response=request.url, redirect_uri=redirect_uri)
                            if token_test:
                                headers = {'Authorization': 'Bearer ' + token_test['access_token']}
                                
                                profile_url = data['userinfo_endpoint']#'https://openidconnect.googleapis.com/v1/userinfo'
                                if profile_url:
                                    userinfo = oaclient.get(profile_url, headers=headers)
                                
                                if userinfo:
                                    userinfo_data = userinfo.json()
                                    print(userinfo_data)
                                    email = userinfo_data['email']
                                    external_id = '{}:{}'.format(auth.name, userinfo_data.get('sub',
                                                                                              userinfo_data.get('id',
                                                                                                                None)))
                                    user = get_or_create_delegated_account(org,
                                                                           'person',
                                                                           'visitor',
                                                                           'Lead',
                                                                           email,
                                                                           email,
                                                                           email,
                                                                           external_id)
                                    if user:
                                        request.query_string = to_bytes(to_unicode(request.query_string) + '&client_id={}'.format(client_id))
                                        return authorization.create_authorization_response(grant_user=user)
                                        return authorization.create_authorization_response(request=request, grant_user=user)
                                    else:
                                        return jsonify(isError=True,
                                                       message="Failure",
                                                       statusCode=401,
                                                       data='Invalid user'), 401
                                else:
                                    return jsonify(isError=True,
                                                   message="Failure",
                                                   statusCode=400,
                                                   data='Cannot retrieve user profile info'), 400
                            else:
                                return jsonify(isError=True,
                                               message="Failure",
                                               statusCode=401,
                                               data='Invalid token'), 401

                        except Exception as e:
                            print('error while fetching token {}'.format(e))
            else:
                return jsonify(isError=True,
                               message="Failure",
                               statusCode=404,
                               data='Authenticator {0} not found'.format(name)), 404
        else:
            return jsonify(isError=True,
                           message="Failure",
                           statusCode=404,
                           data='Client {0} not found'.format(client_id)), 404

    return jsonify(isError=True,
                   message="Failure",
                   statusCode=404,
                   data='Tenant {0} not found'.format(id)), 404

@router_bp.route('/<id>/auth', methods=['GET'])
def tenant_auths(id):
    contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
    org =  org = contextProvider.get_org_by_name(id)
    if org:
        oauths = [{ 'name': n['name'], 'type': 'oauth'} for n in [q.to_dict() for q in contextProvider.get_org_authenticators(org.id)]]
        oauths.append({'name': 'password', 'type': 'password'})
        return jsonify(oauths)
    else:
        return jsonify(isError=True,
                   message="Failure",
                   statusCode=404,
                   data='Tenant {0} not found'.format(id)), 404

@router_bp.route('/', defaults={'endpoint_name': None, 'path': None}, methods=['GET'])
@router_bp.route('/<endpoint_name>', defaults={'path': None}, methods=['GET', 'POST'])
@router_bp.route('/<endpoint_name>/<path:path>', methods=['GET', 'POST'])
@require_oauth(optional=True)
def router(endpoint_name, path):
    contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']

    account = None
    
    content_type = 'text/html'

    if request.accept_mimetypes.accept_json:
        content_type = 'application/json'
    elif 'application/xml' in request.accept_mimetypes:
        content_type = 'application/xml'
    
    if not current_token:
        account = current_user(contextProvider)
        if not account and endpoint_name != 'signin' and endpoint_name != 'public':
            if content_type == 'application/json':
                return jsonify(isError=True,
                        message="Failure",
                        statusCode=401,
                        data='Unauthorized'.format(id)), 401
            else:
                return redirect('/signin')
    else:
        account = current_token.account

    if not account and endpoint_name == 'public':
        orgs = contextProvider.get_public_orgs()
        if path == 'apps':
            apps = []
            for org in orgs:
                apps = apps + [a for a in contextProvider.get_public_apps(org.id)]
            return jsonify([app.to_dict() for app in apps])
        elif path == 'types':
            types = []
            for org in orgs:
                types = types + [t for t in contextProvider.get_public_types(org.id)]
            return jsonify([type.to_dict() for type in types])
        elif path == 'forms':
            forms = []
            for org in orgs:
                forms = forms + [f for f in contextProvider.get_public_forms(org.id)]
            return jsonify([form.to_dict() for form in forms])
        elif path and path.startswith('items/'):
            parts = path.split('/')
            subpath = parts[0]
            if subpath == 'items':
                id = None
                if len(parts) > 1:
                    id = parts[1]
                if id:
                    item = contextProvider.get_public_item(id)
                    if item:
                        return jsonify(item.to_dict())
                    else:
                        return jsonify(isError=True,
                                       message="Failure",
                                       statusCode=404,
                                       data='Item {0} not found'.format(id)), 404
            else:
                print(path)

    context = contextProvider.context(account.org.id, account.id)

    if endpoint_name == 'signin':
        return render_template('login.html')
    elif not endpoint_name:
        endpoints = context.get_endpoints(context)
        endpoint = next((e for e in endpoints), None)
        endpoint_name = endpoint.item.name.lower()
        return redirect('/{}'.format(endpoint_name))
    
    

    endpoint = context.get_endpoint(context, endpoint_name)
    method = request.method.lower()

    if endpoint:
        args = request.args.to_dict()

        if request.method.lower() == 'post' or request.method.lower() == 'put':
            args |= request.form.to_dict()
            if args:
                json_args = request.get_json(silent=True)
                if json_args:
                    args |= json_args
            else:
                args = request.get_json(silent=True)

        if '_method' in args:
            method = args['_method'].lower() 

        return endpoint.invoke(context, method, args, path=path, content_type=content_type)
    else:
        return jsonify(isError=True,
                        message="Failure",
                        statusCode=404,
                        data='Not found'), 404


import importlib.util
import sys

from realnet.core.provider import ResourceProvider
from realnet.core.type import Func


class GenericResourceProvider(ResourceProvider):

    def __init__(self):
        pass

    def get_resources(self, module):
        # account = module.get_account()
        return [] #[Endpoint(e) for e in module.find_items({'types': ['Endpoint'], 'children': 'true'}) if self.acl.can_account_read_item(account, e)]

    def get_resource_method(self, module, resource_name, method_name):
        account = module.get_account()
        resources = [r for r in module.find_items({ 'types': ['Resource'], 
                                                    'name': resource_name, 
                                                    'any_level': 'true',
                                                    'keys': ['module'], 
                                                    'values': ['true']}) if module.can_account_read_item(account, r)]
        resource_item = next(iter(resources), None)
        if resource_item:
            # Check if there is a child function with that name and return that
            funcs = [f for f in resource_item.items_of_type('Func') if f.name.lower() == method_name.lower()]
            if funcs:
                func = next(iter(funcs), None)
                if func:
                    return Func(func)

            module_path = resource_item.attributes.get('module_path')
            module_class_name = resource_item.attributes.get('module_class')
            if module_path:
                resource_module = importlib.import_module(module_path)
                resource_class = getattr(resource_module, module_class_name)
                resource_instance = resource_class()
                func = getattr(resource_instance, method_name, None)
                if callable(func):
                    return Func(callback=lambda module,args,path,content_type : func(module, args, path, content_type))

            code = resource_item.attributes.get('code')
            if not code:
                resource_data = module.get_data(resource_item.id)
                if resource_data:
                    code = resource_data.decode('utf-8')
                    return
            
        return None


from ..runner import Runner
from .app import create_app
from realnet.core.config import Config

class HttpRunner(Runner):
    
    def run(self, context_provider):
        app = create_app(context_provider)
        cfg = Config()
        app.run(cfg.get_server_host(), cfg.get_server_port())
        



from flask import render_template
from realnet.resource.items.items import Items

class Roles(Items):
    
    def get_endpoint_name(self):
        return 'roles'

    def get_items(self, module, account, query, parent_item=None):
        roles_query = {'types': ['Role']}
        return [i for i in module.find_items(roles_query) if module.can_account_read_item(account, i)]

    def post(self, module, args, path=None, content_type='text/html'):
        if path == 'apps':
            params = dict()
            parent_id = args.get('id', None)
            


        module.create_item(**args)
        del args['add']
        return self.render_item(module, args, path, content_type)

    def put(self, module, args, path=None, content_type='text/html'):
        params = dict()
        if 'name' in args:
            params['name'] = args['name']
        if 'attributes' in args:
            params['attributes'] = args['attributes']
        module.update_item(args['id'], **params)
        del args['id']
        del args['edit']
        del args['item_id']
        return self.render_item(module, args, path, content_type)

    def delete(self, module, args, path=None, content_type='text/html'):
        module.delete_item(args['id'])
        del args['id']
        del args['delete']
        del args['item_id']
        return self.render_item(module, args, path, content_type)


import mimetypes

from flask import jsonify
from realnet.resource.items.items import Items

class Files(Items):
    
    def get_endpoint_name(self):
        return 'files'

    def get(self, module, args, path=None, content_type='text/html'):
        if path == 'upload-url':
            attributes = dict()
            attributes['filename'] = args.get('filename', 'file.tmp')
            attributes['filesize'] = args.get('size', 0)
            content_type = mimetypes.guess_type(attributes['filename'].lower())[0]
            if not content_type:
                if attributes['filename'].lower().endswith('heic'):
                    content_type = 'image/heic'
                else:
                    content_type = 'application/octet-stream'
            
            arguments = dict()
            typename = 'File'
            if content_type.startswith('image/'):
                typename = 'Image'
            elif content_type.startswith('application/pdf') or content_type.startswith('text/plain') or attributes['filename'].endswith('.md'):
                typename = 'Document'
                if attributes['filename'].endswith('.md'):
                    content_type = 'text/markdown'
            elif content_type.startswith('text/html'):
                typename = 'Page'
            elif content_type.startswith('video/'):
                typename = 'Video'
            elif content_type.startswith('application/zip') or content_type.startswith('application/x-zip-compressed'):
                typename = 'File'
            elif content_type.startswith('text/csv'):
                typename = 'File'
            elif attributes['filename'].endswith('.glb'):
                typename = 'Scene'
            
            attributes['content_type'] = content_type
            arguments['type'] = args.get('type', typename)
            parent_id = args.get('parent_id', None)
            if parent_id:
                parent_item = module.get_item(parent_id)
                if parent_item and module.can_account_write_item(module.get_account(), parent_item):
                    arguments['parent_id'] = parent_id
            
            arguments['ip_address'] = args.get('ip_address', None)
            for key, value in args.items():
                if not key in {'filename', 'file_size', 'parent_id', 'type', 'size', 'ip_address'}:
                    attributes[key] = value
            
            arguments['name'] = attributes['filename']
            arguments['attributes'] = attributes
            item = module.create_item(**arguments)
            if item:
                return module.get_data_upload_url(item.id)

        return self.render_item(module, args, path, content_type)

    def post(self, module, args, path=None, content_type='text/html'):
        if path == 'upload-confirm':
            item_id = args.get('item_id', None)
            filename = args.get('filename', None)
            filesize = args.get('size', None)
            content_type = mimetypes.guess_type(filename.lower())[0]

            if not content_type:
                if filename.lower().endswith('heic'):
                    content_type = 'image/heic'
                else:
                    content_type = 'application/octet-stream'
            
            item = module.get_item(item_id)
            if item:
                # todo get from S3
                attributes = dict(item.attributes)
                attributes['filename'] = filename
                attributes['content_type'] = content_type
                attributes['content_length'] = filesize
                attributes['mime_type'] = content_type
                
                for (key,value) in args.items():
                    if not key in {'filename', 'file_size', 'parent_id', 'type', 'size', 'ip_address'}:
                        attributes[key] = value

                module.update_item(item.id, **{"attributes": attributes})

                return jsonify(item.to_dict()), 200        
            # module.create_item(**args)
        if 'add' in args:
            del args['add']
        return self.render_item(module, args, path, content_type)


from abc import abstractmethod
from .cmd import Cmd

import argparse


class Shell(Cmd):

    def run(self, args=None):

        if args is None:
            parser = argparse.ArgumentParser(prog=self.name)
            self.add_arguments(parser)

            subparsers = parser.add_subparsers(dest='{0}_command'.format(self.name), help=self.get_help())

            for c in self.get_commands():
                c.setup(subparsers)
            args = parser.parse_args()

        if hasattr(args, '{0}_command'.format(self.name)):
            command_name = getattr(args, '{0}_command'.format(self.name))
            # print('looking for command {0}'.format(command_name))
            command = next(iter([c for c in self.get_commands() if c.name == command_name]), None)
            if command:
                return command.run(args)
            else:
                print("{0} is not a valid command".format(command_name))
                return None
        else:
            print('Unable to find a valid command')
            return None




    @abstractmethod
    def get_commands(self):
        return []

    def setup(self, subparsers, parent=None):
        parser = super().setup(subparsers)
        commands = self.get_commands()
        if commands:
            subsubparsers = parser.add_subparsers(dest='{0}_command'.format(self.name))
            for c in commands:
                c.setup(subsubparsers)

    def __call__(self, args):
        parser = argparse.ArgumentParser(prog=self.name)

        self.add_arguments(parser)
        subparsers = parser.add_subparsers(dest='{0}_command'.format(self.name), help=self.get_help())

        for c in self.get_commands():
            c.setup(subparsers)

        return self.run(parser.parse_args(args))



from abc import ABC, abstractmethod
from .type import *

class TypeProvider(ABC):
    @abstractmethod
    def get_types(self):
        pass

    @abstractmethod
    def get_derived_types(self, type_ids):
        pass

    @abstractmethod
    def get_type_by_id(self, id):
        pass

    @abstractmethod
    def get_type_by_name(self, name):
        pass

    @abstractmethod
    def get_type_instances(self, id):
        pass

    @abstractmethod
    def delete_type(self, id):
        pass

    @abstractmethod
    def update_type(self, id, **kwargs):
        pass

    @abstractmethod
    def create_type(self, **kwargs):
        pass

    @abstractmethod
    def create_instance(self, **kwargs):
        pass

class ItemProvider(ABC):
    
    @abstractmethod
    def get_items(self):
        pass

    @abstractmethod
    def get_item(self, id, children=False):
        pass

    @abstractmethod
    def delete_item(self, id):
        pass

    @abstractmethod
    def update_item(self, id, **kwargs):
        pass

    @abstractmethod
    def create_item(self, **kwargs):
        pass

    @abstractmethod
    def find_items(self, query):
        pass

class DataProvider(ABC):
    
    @abstractmethod
    def get_data(self, id):
        pass

    @abstractmethod
    def update_data(self, id, storage):
        pass

    @abstractmethod
    def delete_data(self, id):
        pass

    @abstractmethod
    def get_data_upload_url(self, id):
        pass

    @abstractmethod
    def confirm_data_upload(self, id):
        pass

class AclProvider(ABC):
    
    @abstractmethod
    def is_item_public(self, item):
        pass

    @abstractmethod
    def can_account_execute_item(self, account, item):
        pass

    @abstractmethod
    def can_account_message_item(self, account, item):
        pass

    @abstractmethod
    def can_account_read_item(self, account, item):
        pass

    @abstractmethod
    def can_account_write_item(self, account, item):
        pass

    @abstractmethod
    def can_account_delete_item(self, account, item):
        pass

class AppProvider(ABC):
    
    @abstractmethod
    def get_apps(self, module):
        pass

class RolesProvider(ABC):
    
    @abstractmethod
    def get_roles(self, module):
        pass

    @abstractmethod
    def create_role(self, **kwargs):
        pass

    @abstractmethod
    def delete_role(self, id):
        pass

    @abstractmethod
    def update_role(self, id, **kwargs):
        pass

class OrgProvider(ABC):
    
    @abstractmethod
    def get_account_groups(self, account_id):
        pass

    @abstractmethod
    def get_org_groups(self, org_id):
        pass


class GroupProvider(ABC):
    
    @abstractmethod
    def get_groups(self):
        pass

class AccountProvider(ABC):
    
    @abstractmethod
    def get_account(self):
        pass

    @abstractmethod
    def get_org(self):
        pass

class TopicProvider(ABC):
    
    @abstractmethod
    def get_topics(self):
        pass

    @abstractmethod
    def get_topic(self, topic_name):
        pass

class EndpointProvider(ABC):
    
    @abstractmethod
    def get_endpoints(self, module):
        pass

    @abstractmethod
    def get_endpoint(self, module, endpoint_name):
        pass

class ResourceProvider(ABC):
    
    @abstractmethod
    def get_resources(self, module):
        pass

    @abstractmethod
    def get_resource_method(self, module, resource_name, method_name):
        pass


class ModuleProvider(ABC):
    
    @abstractmethod
    def get_modules(self):
        pass

    @abstractmethod
    def get_module(self, module_name):
        pass

class ClientProvider(ABC):
    
    @abstractmethod
    def get_clients(self):
        pass

    @abstractmethod
    def get_client(self, client_id):
        pass

class ResponseProvider(ABC):
    
    @abstractmethod
    def render_template(self, **kwargs):
        pass

    @abstractmethod
    def render_template_string(self, **kwargs):
        pass

    @abstractmethod
    def make_response(self, **kwargs):
        pass

    @abstractmethod
    def send_file(self, **kwargs):
        pass

class GroupAclProvider(AclProvider):

    def __init__(self, groups):
        self.groups = groups

    def is_item_public(self, item):
        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True
        return False

    def can_account_execute_item(self, account, item):
        
        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and acl.permission and  'e' in acl.permission]:
            return True

        account_groups = self.groups.get_groups(account)
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.id)

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and  acl.permission and 'e' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and  acl.permission and 'e' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_message_item(self, account, item):

        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and  acl.permission and 'm' in acl.permission]:
            return True

        account_groups = self.groups.get_groups(account)
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.id)

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  'm' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and acl.permission and  'm' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_read_item(self, account, item):

        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and acl.permission and ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        account_groups = self.groups.get_groups(account)
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.id)

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_write_item(self, account, item):

        if [acl for acl in item.acls if
            acl.type == AclType.user and acl.target_id == account.id and acl.permission and 'w' in acl.permission]:
            return True

        account_groups = self.groups.get_groups(account)
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.id)

        if [acl for acl in item.acls if
            acl.type == AclType.group and acl.target_id in group_ids and acl.permission and 'w' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  'w' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True

    def can_account_delete_item(self, account, item):
        return item.owner_id == account.id and account.org_id == item.org_id

class OrgsProvider(ABC):    

    @abstractmethod
    def get_orgs(self):
        pass

    @abstractmethod
    def get_org_by_id(self,id):
        pass

    @abstractmethod
    def get_org_by_name(self,name):
        pass

    @abstractmethod
    def get_account_by_id(self,id):
        pass

    @abstractmethod
    def check_password(self, org_id, account_id, password):
        pass

    @abstractmethod
    def get_org_authenticators(self, org_id):
        pass

    @abstractmethod
    def get_org_clients(self, org_id):
        pass

    @abstractmethod
    def get_org_client(self, org_id, client_id):
        pass

    @abstractmethod
    def get_public_apps(self, org_id):
        pass

    @abstractmethod
    def get_public_types(self, org_id):
        pass

    @abstractmethod
    def get_public_forms(self, org_id):
        pass

    @abstractmethod
    def get_public_orgs(self):
        pass

    @abstractmethod
    def get_public_item(self, id):
        pass

class InitializationProvider(ABC):    

    @abstractmethod
    def initialize(self, org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri):
        pass

class ImportProvider(ABC):

    @abstractmethod
    def import_structure(self, module, path):
        pass

class ContextProvider(OrgsProvider, InitializationProvider):
    
    @abstractmethod
    def context(self, org_id, account_id):
        pass
    

class Module( TypeProvider, 
              ItemProvider, 
              DataProvider, 
              AclProvider,
              AppProvider,
              OrgProvider,
              GroupProvider,
              TopicProvider,
              EndpointProvider,
              ModuleProvider,
              AccountProvider,
              ResourceProvider,
              ImportProvider,
              ClientProvider):
    pass

class Context(Module):
    def __init__(   self, 
                    types, 
                    items, 
                    data, 
                    apps, 
                    orgs, 
                    groups, 
                    topics, 
                    endpoints, 
                    modules, 
                    accounts, 
                    acls,
                    resources,
                    importer,
                    client):
        self.types = types
        self.items = items
        self.data = data
        self.apps = apps
        self.orgs = orgs
        self.groups = groups
        self.topics = topics
        self.endpoints = endpoints
        self.modules = modules
        self.accounts = accounts
        self.acls = acls
        self.resources = resources
        self.importer = importer
        self.client = client

    def get_types(self):
        return self.types.get_types()

    def get_derived_types(self, type_ids):
        return self.types.get_derived_types(type_ids)

    def get_type_by_id(self, id):
        return self.types.get_type_by_id(id)

    def get_type_by_name(self, name):
        return self.types.get_type_by_name(name)

    def get_type_instances(self, id):
        return self.types.get_type_instances(id)

    def delete_type(self, id):
        return self.types.delete_type(id)

    def update_type(self, id, **kwargs):
        return self.types.update_type(id, **kwargs)

    def create_type(self, **kwargs):
        return self.types.create_type(**kwargs)

    def create_instance(self, **kwargs):
        return self.types.create_instance(**kwargs)

    def get_items(self):
        return self.items.get_items()

    def get_item(self, id, children=False):
        return self.items.get_item(id, children)

    def delete_item(self, id):
        return self.items.delete_item(id)

    def update_item(self, id, **kwargs):
        return self.items.update_item(id, **kwargs)

    def create_item(self, **kwargs):
        return self.items.create_item(**kwargs)

    def find_items(self, query):
        return self.items.find_items(query)

    def get_data(self, id):
        return self.data.get_data(id)

    def update_data(self, id, storage):
        return self.data.update_data(id, storage)

    def delete_data(self, id):
        return self.data.delete_data(id)

    def get_data_upload_url(self, id):
        return self.data.get_data_upload_url(id)

    def confirm_data_upload(self, id):
        return self.data.confirm_data_upload(id)

    def is_item_public(self, item):
        return self.acls.is_item_public(item)

    def can_account_execute_item(self, account, item):
        return self.acls.can_account_execute_item(account, item)


    def can_account_message_item(self, account, item):
        return self.acls.can_account_message_item(account, item)


    def can_account_read_item(self, account, item):
        return self.acls.can_account_read_item(account, item)


    def can_account_write_item(self, account, item):
        return self.acls.can_account_write_item(account, item)

    def can_account_delete_item(self, account, item):
        return self.acls.can_account_delete_item(account, item)

    def get_apps(self, module):
        return self.apps.get_apps(module)

    def get_account_groups(self, account_id):
        return self.orgs.get_account_groups(self, account_id)

    def get_org_groups(self, org_id):
        return self.orgs.get_org_groups(org_id)

    def get_groups(self):
        return self.groups.get_groups()

    def get_topics(self):
        return self.topics.get_topics()

    def get_topic(self, topic_name):
        return self.topics.get_topic(topic_name)

    def get_endpoints(self, module):
        return self.endpoints.get_endpoints(module)

    def get_endpoint(self, module, endpoint_name):
        return self.endpoints.get_endpoint(module, endpoint_name)

    def get_modules(self):
        return self.modules.get_modules()

    def get_module(self, module_name):
        return self.modules.get_module(module_name)

    def get_account(self):
        return self.accounts.get_account()

    def get_org(self):
        return self.orgs.get_org()

    def get_resources(self, module):
        return self.resources.get_resources(module)

    def get_resource_method(self, module, resource_name, method_name):
        return self.resources.get_resource_method(module, resource_name, method_name)

    def import_structure(self, module, path):
        return self.importer.import_structure(module, path)

    def get_clients(self):
        return self.client.get_clients()

    def get_client(self, client_id):
        return self.client.get_client(client_id)
        





import enum

class AclType(enum.Enum):
    public = 1
    group = 2
    user = 3
    org = 4

class Acl:
    
    def __init__(self, type, permission, target_id, org_id, owner_id):
        self.type = type
        self.permission = permission
        self.target_id = target_id
        self.org_id = org_id
        self.owner_id = owner_id


from realnet.shell import ProtoCmd

class Info(ProtoCmd):
    
    def __init__(self):
        super().__init__('info',
                         'realnet info')

    def add_arguments(self, parser):
        pass

    def run(self, args):
        print('this is info')


import os
from dotenv import *

path = os.path.join(os.getcwd(), ".env")
if os.path.exists(path):
    load_dotenv(dotenv_path=path)

class Config:

    def get_database_url(self):
        if(os.getenv('REALNET_DB_TYPE') == 'sqlite'):
            return '{0}:///{1}.db'.format(os.getenv('REALNET_DB_TYPE'),
                                                         os.getenv('REALNET_DB_NAME'))
        else:
            return '{0}://{1}:{2}@{3}:{4}/{5}'.format(os.getenv('REALNET_DB_TYPE'),
                                                         os.getenv('REALNET_DB_USER'),
                                                         os.getenv('REALNET_DB_PASS'),
                                                         os.getenv('REALNET_DB_HOST'),
                                                         os.getenv('REALNET_DB_PORT'),
                                                         os.getenv('REALNET_DB_NAME'))

    def get_db_type(self):
        return os.getenv('REALNET_DB_TYPE')

    def get_server_host(self):
        return os.getenv('REALNET_SERVER_HOST')

    def get_server_port(self):
        return os.getenv('REALNET_SERVER_PORT')

    def get_storage_type(self):
        return os.getenv('REALNET_STORAGE_TYPE')

    def get_storage_path(self):
        return os.getenv('REALNET_STORAGE_PATH')

    def get_s3_region(self):
        return os.getenv('REALNET_STORAGE_S3_REGION')

    def get_s3_bucket(self):
        return os.getenv('REALNET_STORAGE_S3_BUCKET')

    def get_s3_key(self):
        return os.getenv('REALNET_STORAGE_S3_KEY')

    def get_s3_secret(self):
        return os.getenv('REALNET_STORAGE_S3_SECRET')

    def get_app_secret(self):
        return os.getenv('REALNET_APP_SECRET')

    def get_jwt_key(self):
        return os.getenv('REALNET_JWT_KEY')

    def get_jwt_issuer(self):
        return os.getenv('REALNET_JWT_ISSUER')

    def get_base64_encode_data(self):
        return os.getenv('REALNET_BASE64_ENCODE_DATA', False)

    def get_sqs_url(self):
        return os.getenv('REALNET_SQS_URL')

    def get_use_s3_upload_url(self):
        return os.getenv('REALNET_USE_S3_UPLOAD_URL', False)



from flask import render_template, jsonify
from realnet.core.type import Resource

class Items(Resource):

    def render_items_json(self, module, args, path):
        account = module.get_account()
        query = self.get_query(module, args, path)
        if path:
            item = self.get_item(module, account, path)
            if item:
                return jsonify(item.to_dict())
            else:
                return {}
        else:        
            items = self.get_items(module, account, query)
            return jsonify([item.to_dict() for item in items])

    def render_items_html(self, module, args, path=None):
        account = module.get_account()
        org = module.get_org()
        apps = module.get_apps(module)
        forms = []
        
        active_app_name = self.get_endpoint_name()
        active_view_name = args.get('view')
        active_subview_name = args.get('subview')

        app = next((a for a in apps if a.name.lower() == self.get_endpoint_name()), None)
        
        if apps:
            if active_app_name:
                app = next((a for a in apps if a.name.lower() == active_app_name), app)

        path_item = app
        target_item = app

        path_segments = []

        path_item_id = None
        target_item_id = None

        if path:
            path_segments = path.split('/')

        if path_segments:
            path_item_id = path_segments[0]
            path_item =  module.get_item(path_item_id)
            target_item = path_item
        
        if 'item_id' in args:
            target_item_id = args['item_id']
            target_item = module.get_item(target_item_id)

        if path_item:
            app = path_item


        views = [i for i in app.items if i.instance.type.is_derived_from('View')]

        items = []
        query = app.attributes.get('query')
        types = app.attributes.get('types',[])
        
        menu = next((i for i in app.items if i.instance.type.is_derived_from('Menu')),None)

        form = None
        

        #     form = next(iter([f for f in module.find_items({'keys': ['type'], 'values': ['App'], 'types': ['Form'], 'children': 'true'}) if module.can_account_read_item(account, f)]), None)
        # elif args.get('delete') == 'true':
        #    pass

        if views:
            active_view = None
            active_subview = None
            if not active_view_name:
                active_view = next((v for v in views), None)
                if active_view:
                    active_view_name = active_view.name.lower()
            else:
                active_view = next((v for v in views if v.name.lower() == active_view_name), None)
            
            if active_view:
                view_query = active_view.attributes.get('query')
                if view_query:
                    query = view_query

                view_types = active_view.attributes.get('types')
                if view_types:
                    types = view_types

                subviews = [i for i in active_view.items if i.instance.type.is_derived_from('View')]
                if not active_subview_name:
                    active_subview = next((v for v in subviews), None)
                    if active_subview:
                        active_subview_name = active_subview.name
                else:
                    active_subview = next((v for v in subviews if v.name == active_subview_name), None)

            if active_subview:
                subview_query = active_subview.attributes.get('query')
                if subview_query:
                    query = subview_query

                subview_types = active_subview.attributes.get('types')
                if subview_types:
                    types = subview_types

        keys = { 'keys': ['type' for type in types ]  }

        if args.get('add') == 'true':
            forms = [f for f in module.find_items({'keys':  ['type' for type in types ], 'values': [type for type in types ], 'types': ['CreateForm'], 'any_level': 'true', 'op': 'or'}) if module.can_account_read_item(account, f)]
        elif args.get('edit') == 'true':
            forms = [f for f in module.find_items({'keys':  ['type' for type in types ], 'values': [type for type in types ], 'types': ['EditForm'], 'any_level': 'true', 'op': 'or'}) if module.can_account_write_item(account, f)]
        elif args.get('delete') == 'true':
            forms = [f for f in module.find_items({'types': ['DeleteForm'], 'any_level': 'true'}) if module.can_account_write_item(account, f)]
        
        items = self.get_items(module, account, query, path_item)

        root_path = '/' + self.get_endpoint_name()

        if path_item_id:
            root_path = root_path + '/' + path_item_id

        if active_view_name:
            root_path = '?view={}&'.format(active_view_name)
        else:
            root_path = root_path + '?'

        typenames = []

        if types:
            type_filter = set(types)
            tbn = module.get_types()
            types_by_id = {t.id:t for t in tbn}
            type_ids = set([t.id for t in tbn if t.name in type_filter])
            typenames = [types_by_id[t].name for t in module.get_derived_types(type_ids)]

        return render_template('item.html', 
                                app=app,
                                item=target_item,
                                apps=apps, 
                                org=org,
                                account=account,
                                items=items, 
                                views=views, 
                                active_view_name=active_view_name, 
                                active_app_name=active_app_name, 
                                active_subview_name = active_subview_name, 
                                types=types,
                                menu=menu,
                                forms=forms,
                                typenames=typenames,
                                root_path=root_path)

    def render_item(self, module, args, path=None, content_type='text/html'):
        if content_type == 'application/json':
            return self.render_items_json(module, args, path)
        else:
            return self.render_items_html(module, args, path)
        

    def get(self, module, args, path=None, content_type='text/html'):
        return self.render_item(module, args, path, content_type)

    def post(self, module, args, path=None, content_type='text/html'):
        module.create_item(**args)
        del args['add']
        return self.render_item(module, args, path, content_type)

    def put(self, module, args, path=None, content_type='text/html'):
        params = dict()
        if 'name' in args:
            params['name'] = args['name']
        if 'attributes' in args:
            params['attributes'] = args['attributes']
        module.update_item(args['id'], **params)
        del args['id']
        del args['edit']
        del args['item_id']
        return self.render_item(module, args, path, content_type)

    def delete(self, module, args, path=None, content_type='text/html'):
        module.delete_item(args['id'])
        del args['id']
        del args['delete']
        del args['item_id']
        return self.render_item(module, args, path, content_type)

    def message(self, module, args, path=None, content_type='text/html'):
        pass

    def run(self, module, args, path=None, content_type='text/html'):
        pass

    def get_data(self, module, args, path=None, content_type='text/html'):
        pass

    def update_data(self, module, args, path=None, content_type='text/html'):
        pass

    def delete_data(self, module, args, path=None, content_type='text/html'):
        pass

    def get_endpoint_name(self):
        return 'items'

    def get_items(self, module, account, query, parent_item=None):
        if query:
            external_types = []
            internal_types = []
            if 'types' in query:
                tbn = {t.name: t for t in module.get_types()}
                if isinstance(query['types'], str):
                    type = tbn.get(query['types'],None)
                    if type:
                        if type.attributes and 'resource' in type.attributes:
                            external_types.append(type)
                        else:
                            internal_types.append(type)
                else:
                    for typename in query['types']:
                        type = tbn.get(typename, None)
                        if type:
                            if type.attributes and 'resource' in type.attributes:
                                external_types.append(type)
                            else:
                                internal_types.append(type)
                if external_types:
                    results = []
                    external_query = dict()
                    for k,v in query.items():
                        if k != 'types' and k != 'type_names':
                            external_query[k] = v

                    if parent_item and 'children' in external_query and external_query['children'] == 'true':
                        external_query['parent_id'] = parent_item.id
                    
                    for external_type in external_types:
                        func = module.get_resource_method(module, external_type.attributes['resource'], 'get_items')
                        if func:
                            results = results + func.invoke(module, account, external_query, parent_item)
                    
                    if internal_types:
                        external_query['types'] = [t.name for t in internal_types]
                        results = results +  [i for i in module.find_items(external_query) if module.can_account_read_item(account, i)]
                    return results
            
            if parent_item and 'children' in query and query['children'] == 'true':
                query['parent_id'] = parent_item.id
            return [i for i in module.find_items(query) if module.can_account_read_item(account, i)]

        return []

    def get_item(self, module, account, path):
        item = module.get_item(path)
        if item and module.can_account_read_item(account, item):
            return item

        return None

    def get_query(self, module, args, path):
        query = dict()
        
        if 'types' in args:
            if isinstance(args['types'], str):
                query['types'] = [args['types']]
            else:
                query['types'] = args['types']

        if 'any_level' in args:
            query['any_level'] = args['any_level']
        
        return query

    def get_types(self, module, account, query, parent_item=None):
        return None


from flask import render_template
from realnet.resource.items.items import Items

class Apps(Items):
    
    def get_endpoint_name(self):
        return 'apps'

    def get_query(self, module, account, query, parent_item=None):
        return {'types':['App']}


import enum
import json
import uuid
import time

from sqlalchemy import CheckConstraint, create_engine
import sqlalchemy as db
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import declarative_base, relationship, Session

from sqlalchemy_serializer import SerializerMixin

import shapely

from geoalchemy2 import Geometry
from geoalchemy2.elements import WKBElement
from geoalchemy2.shape import to_shape

from werkzeug.security import generate_password_hash, check_password_hash, gen_salt

from authlib.integrations.sqla_oauth2 import (
    OAuth2ClientMixin,
    OAuth2AuthorizationCodeMixin,
    OAuth2TokenMixin,
)

from realnet.core.config import Config

cfg = Config()

engine = create_engine(cfg.get_database_url(), echo=True)
Model = declarative_base()
session = Session(engine)

class Authenticator(Model, SerializerMixin):
    __tablename__ = "authenticator"
    id = db.Column(db.String(36), primary_key=True)
    name = db.Column(db.String(40))
    api_base_url = db.Column(db.String(128))
    request_token_url = db.Column(db.String(128))
    access_token_url = db.Column(db.String(128))
    authorize_url = db.Column(db.String(128))
    client_kwargs = db.Column(db.JSON())
    client_id = db.Column(db.String(255))
    client_secret = db.Column(db.String(255))
    userinfo_endpoint = db.Column(db.String(128))
    server_metadata_url = db.Column(db.String(128))
    redirect_url = db.Column(db.String(128))
    scope = db.Column(db.String(128))
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)

    def get_url(self):
        return self.authorize_url + '?client_id=' + self.client_id + '&response_type=code' + '&scope=openid profile email' + '&redirect_uri=http://localhost:8080/oauth/authorize'


class AccountType(enum.Enum):
    person = 1
    thing = 2

class OrgRoleType(enum.Enum):
    superuser = 1
    admin = 2
    user = 3
    visitor = 4

class Account(Model, SerializerMixin):
    __tablename__ = "account"
    id = db.Column(db.String(36), primary_key=True)
    type = db.Column(db.Enum(AccountType))
    username = db.Column(db.String(40), unique=True)
    email = db.Column(db.String(254), unique=True)
    password_hash = db.Column(db.String(128))
    attributes = db.Column(db.JSON)
    external_id = db.Column(db.String(254))
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    org_role_type = db.Column(db.Enum(OrgRoleType), nullable=False)
    attributes = db.Column(db.JSON)
    groups = relationship('AccountGroup', back_populates='account')
    roles = relationship('AccountRole', back_populates="account")
    org = relationship('Org', foreign_keys='[Account.org_id]')

    def set_password(self, password):
        self.password_hash = generate_password_hash(password)

    def check_password(self, password):
        # print(password)
        return check_password_hash(self.password_hash, password)

    def get_user_id(self):
        return self.id

    def __str__(self):
        return self.username


class GroupRoleType(enum.Enum):
    superadmin = 1
    admin = 2
    member = 3


# Define the Org data-model
class Org(Model, SerializerMixin):
    __tablename__ = "org"
    id = db.Column(db.String(36), primary_key=True)
    name = db.Column(db.String(50), unique=True)
    attributes = db.Column(db.JSON)
    public = db.Column(db.Boolean, default=False)
    logo = db.Column(db.LargeBinary)
    color = db.Column(db.String(8))

# Define the Group data-model
class Group(Model, SerializerMixin):
    __tablename__ = "group"
    id = db.Column(db.String(36), primary_key=True)
    name = db.Column(db.String(50), unique=True)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id'), nullable=False)
    attributes = db.Column(db.JSON)
    org = relationship('Org', foreign_keys='[Group.org_id]')
    accounts = relationship('AccountGroup', back_populates='group')

# Define the AccountGroup association table
class AccountGroup(Model, SerializerMixin):
    __tablename__ = "account_group"
    id = db.Column(db.String(36), primary_key=True)
    account_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    group_id = db.Column(db.String(36), db.ForeignKey('group.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    role_type = db.Column(db.Enum(GroupRoleType), nullable=False)
    account = relationship('Account')
    group = relationship('Group')

# Define the Role data-model
class Role(Model, SerializerMixin):
    __tablename__ = "role"
    id = db.Column(db.String(36), primary_key=True)
    name = db.Column(db.String(50))
    org_id = db.Column(db.String(36), db.ForeignKey('org.id'), nullable=False)
    attributes = db.Column(db.JSON)
    org = relationship('Org', foreign_keys='[Role.org_id]')
    apps = relationship('RoleApp', back_populates='role')

# Define the AccountRole association table
class AccountRole(Model, SerializerMixin):
    __tablename__ = "account_role"
    id = db.Column(db.String(36), primary_key=True)
    account_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    role_id = db.Column(db.String(36), db.ForeignKey('role.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    attributes = db.Column(db.JSON)
    account = relationship('Account', back_populates="roles")
    role = relationship('Role')

# Define the RoleApp association table
class RoleApp(Model, SerializerMixin):
    __tablename__ = "role_app"
    id = db.Column(db.String(36), primary_key=True)
    role_id = db.Column(db.String(36), db.ForeignKey('role.id', ondelete='CASCADE'), nullable=False)
    app_id = db.Column(db.String(36), db.ForeignKey('item.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    attributes = db.Column(db.JSON)
    role = relationship('Role')
    app = relationship('Item')

class Token(Model, OAuth2TokenMixin):
    __tablename__ = "token"
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    account = relationship('Account')


class AuthorizationCode(Model, OAuth2AuthorizationCodeMixin):
    __tablename__ = "authorization_code"
    id = db.Column(db.Integer, primary_key=True)
    account_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    account = relationship('Account')


class Client(Model, OAuth2ClientMixin, SerializerMixin):
    __tablename__ = "client"
    id = db.Column(db.String(36), primary_key=True)
    name = db.Column(db.String(42))
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    attributes = db.Column(db.JSON)
    org = relationship('Org', foreign_keys='[Client.org_id]')

    def get_allowed_scope(self, scope):
        if not scope:
            return ''
        allowed = set(self.scope)
        scopes = [s for s in self.scope]
        return ' '.join([s for s in scopes if s in allowed])


class Type(Model, SerializerMixin):
    __tablename__ = "type"
    id = db.Column(db.String(36), primary_key=True, index=True)
    name = db.Column(db.String(128))
    icon = db.Column(db.String(128))
    attributes = db.Column(db.JSON)
    owner_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    base_id = db.Column(db.String(36), db.ForeignKey('type.id', ondelete='CASCADE'))
    module = db.Column(db.String(128))
    base = relationship('Type', remote_side='[Type.id]')
    instances = relationship('Instance', foreign_keys='[Instance.parent_type_id]')
    acls = relationship('Acl', passive_deletes=True)

class VisibilityType(enum.Enum):
    visible = 1
    hidden = 2
    restricted = 3

class Instance(Model, SerializerMixin):
    __tablename__ = "instance"
    id = db.Column(db.String(36), primary_key=True, index=True)
    name = db.Column(db.String(128))
    icon = db.Column(db.String(128))
    attributes = db.Column(db.JSON)
    public = db.Column(db.Boolean)
    visibility = db.Column(db.Enum(VisibilityType))
    owner_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    type_id = db.Column(db.String(36), db.ForeignKey('type.id', ondelete='CASCADE'), nullable=False)
    type = relationship('Type', foreign_keys='[Instance.type_id]')
    parent_type_id = db.Column(db.String(36), db.ForeignKey('type.id'))
    acls = relationship('Acl', passive_deletes=True)


def jsonize_geometry(g):
    return json.loads(json.dumps(shapely.geometry.mapping(to_shape(g))))

class Item(Model, SerializerMixin):
    __tablename__ = "item"
    serialize_types = (
        (WKBElement, lambda x: jsonize_geometry(x)),
    )
    id = db.Column(db.String(36), primary_key=True, index=True)
    name = db.Column(db.String(128))
    attributes = db.Column(db.JSON)
    owner_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'), nullable=False)
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    type_id = db.Column(db.String(36), db.ForeignKey('type.id', ondelete='CASCADE'), nullable=False)
    parent_id = db.Column(db.String(36), db.ForeignKey('item.id', ondelete="CASCADE"))
    linked_item_id = db.Column(db.String(36), db.ForeignKey('item.id', ondelete="CASCADE"))
    location = db.Column(Geometry(geometry_type='GEOMETRY', srid=4326))
    valid_from = db.Column(db.DateTime(timezone=False))
    valid_to = db.Column(db.DateTime(timezone=False))
    status = db.Column(db.String(128))
    visibility = db.Column(db.Enum(VisibilityType))
    tags = db.Column(ARRAY(db.String()))
    type = relationship('Type')
    acls = relationship('Acl', passive_deletes=True)
    # items = relationship('Item', primaryjoin='Item.parent_id==Item.id')
    linked_item = relationship('Item', foreign_keys='[Item.linked_item_id]', remote_side='[Item.id]')
    # linked_item = relationship('Item', primaryjoin='Item.id==Item.linked_item_id')
    # parent = relationship('Item', primaryjoin='Item.parent_id==Item.id')


class AclType(enum.Enum):
    public = 1
    group = 2
    user = 3
    org = 4


# Define the Acl data-model
class Acl(Model, SerializerMixin):
    __tablename__ = "acl"
    id = db.Column(db.String(36), primary_key=True)
    type = db.Column(db.Enum(AclType))
    permission = db.Column(db.String(50))
    target_id = db.Column(db.String(36))
    owner_id = db.Column(db.String(36), db.ForeignKey('account.id', ondelete='CASCADE'))
    org_id = db.Column(db.String(36), db.ForeignKey('org.id', ondelete='CASCADE'), nullable=False)
    item_id = db.Column(db.String(36), db.ForeignKey('item.id', ondelete='CASCADE'), nullable=True)
    type_id = db.Column(db.String(36), db.ForeignKey('type.id', ondelete='CASCADE'), nullable=True)
    instance_id = db.Column(db.String(36), db.ForeignKey('instance.id', ondelete='CASCADE'), nullable=True)
    CheckConstraint(
    '(item_id IS NOT NULL AND type_id IS NULL AND instance_id IS NULL) OR' +
    '(item_id IS NULL AND type_id IS NOT NULL AND instance_id IS NULL) OR ' +
    '(item_id IS NULL AND type_id IS NULL AND instance_id IS NOT NULL)', 
    name='thing_id_check')

    @hybrid_property
    def thing_id(self):
        return self.item_id or self.type_id or self.instance_id

def create_client(name,
               uri,
               grant_types,
               redirect_uris,
               response_types,
               scope,
               auth_method,
               org_id,
               client_id=gen_salt(24),
               client_secret=gen_salt(48)):
    client_id_issued_at = int(time.time())
    app_id = str(uuid.uuid4())
    client = Client(
        id=app_id,
        name=name,
        client_id=client_id,
        client_id_issued_at=client_id_issued_at,
        org_id=org_id
    )

    client_metadata = {
        'client_name': name,
        'client_uri': uri,
        'grant_types': grant_types,
        'redirect_uris': redirect_uris,
        'response_types': response_types,
        'scope': scope,
        'token_endpoint_auth_method': auth_method
    }
    client.set_client_metadata(client_metadata)

    if client_metadata['token_endpoint_auth_method'] == 'none':
        client.client_secret = ''
    else:
        client.client_secret = client_secret

    session.add(client)
    session.commit()

    return client

def create_account(org,
                   account_type,
                   account_role,
                   account_username,
                   account_password,
                   account_email,
                   account_external_id):
    id = str(uuid.uuid4())
    account = Account( id=id, 
                            type=account_type,
                            username=account_username, 
                            email=account_email, 
                            org_id=org.id,
                            org_role_type=account_role,
                            external_id=account_external_id)

    if account_password:
        account.set_password(account_password)

    session.add(account)

    return account

def get_or_create_delegated_account(org,
                                    account_type,
                                    account_role,
                                    account_business_role,
                                    account_username,
                                    account_password,
                                    account_email,
                                    account_external_id):
    id = str(uuid.uuid4())

    account = session.query(Account).filter(Account.external_id == account_external_id, Account.org_id == org.id).first()

    if account:
        return account

    account = create_account(   org,
                                AccountType[account_type],
                                OrgRoleType[account_role],
                                account_username,
                                None,
                                account_email,
                                account_external_id)

    session.add(account)

    role = Role.query.filter(Role.org_id == org.id, Role.name == account_business_role).first()
    if role:
        ar = AccountRole(id=str(uuid.uuid4()), org_id=org.id, account_id=account.id, role_id=role.id)
        session.add(ar)
        session.commit()
    
    return account

def initialize(org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri):
    Model.metadata.create_all(engine)

    org = Org(id=str(uuid.uuid4()), name=org_name, public=True)
    session.add(org)
    session.commit()

    root_account_id = str(uuid.uuid4())
    root_account = Account( id=root_account_id, 
                            type=AccountType.person,
                            username=admin_username, 
                            email=admin_email, 
                            org_id=org.id,
                            org_role_type=OrgRoleType.superuser)

    root_account.set_password(admin_password)

    session.add(root_account)

    cli_client_id = gen_salt(24)
    cli_client = create_client(name=org_name + '_cli',
                        client_id=cli_client_id,
                        uri=uri,
                        grant_types=['password'],
                        redirect_uris=[],
                        response_types=['token'],
                        scope='',
                        auth_method='client_secret_basic',
                        org_id=org.id)

    web_client_id = 'IEmf5XYQJXIHvWcQtZ5FXbLM' #gen_salt(24)
    web_client = create_client(name=org_name + '_realscape_web',
                        client_id=web_client_id,
                        uri=uri,
                        grant_types=['password'],
                        redirect_uris=[redirect_uri],
                        response_types=['token'],
                        scope='',
                        auth_method='none',
                        org_id=org.id)

    mobile_client_id = gen_salt(24)
    mobile_client_secret = gen_salt(48)
    mobile_client = create_client(name=org_name + '_realscape_mob',
                            client_id=mobile_client_id,
                            client_secret=mobile_client_secret,
                            uri=uri,
                            grant_types=['authorization_code','password'],
                            redirect_uris=[mobile_redirect_uri],
                            response_types=['code'],
                            scope='',
                            auth_method='client_secret_basic',
                            org_id=org.id)
    return root_account



from realnet.shell import Shell

from realnet.cmd import Info, Get, Create, Runner, Server

class Realnet(Shell):
    
    def __init__(self):
        super().__init__('realnet')

    def get_commands(self):
        return [Info(), Get(), Create(), Runner(), Server()]

    def add_arguments(self, parser):
        pass

    def get_help(self):
        return 'realnet'


import uuid

from realnet.provider.sql.models import Type as TypeModel, Instance as InstanceModel, Item as ItemModel, Acl as AclModel, session as db
from realnet.core.type import Type, Instance, Item
from realnet.core.acl import AclType, Acl


def get_types_by_name(org_id):
    type_models = db.query(TypeModel).filter(TypeModel.org_id == org_id).all()
    tmbn = {tm.name:tm for tm in type_models}
    tbn = dict()
    for tm in type_models:
        type_model_to_type(org_id, tm, tmbn, tbn)
    return tbn

def fill_derived_types(base_types, type_ids, results):
    for type_id in type_ids:
        derived_types = base_types.get(type_id)
        if derived_types:
            for derived_type in derived_types:
                results.add(derived_type)
                fill_derived_types(base_types, [derived_type], results)

def get_derived_types(org_id, type_ids):
    type_models = db.query(TypeModel).filter(TypeModel.org_id == org_id).all()
    base_types = dict()
    
    for tm in type_models:
        if tm.base_id:
            existing_base_type = base_types.get(tm.base_id)
            if existing_base_type:
                existing_base_type.append(tm.id)
            else:
                base_types[tm.base_id] = [tm.id]

    results = set()
    fill_derived_types(base_types, type_ids, results)

    return list(results)

def type_model_to_type(org_id, type_model, type_models_by_name, types_by_name):
    base = type_model.base
    attributes = dict(type_model.attributes or dict())
    if base:
        if base.name in types_by_name:
            base = types_by_name[base.name]
        else:
            base = type_model_to_type(org_id, base, type_models_by_name, types_by_name)
        attributes = base.attributes | attributes
    
    instance_models = db.query(InstanceModel).filter(InstanceModel.parent_type_id == type_model.id, InstanceModel.org_id == org_id).all()
    instances = [instance_model_to_instance(im, types_by_name) for im in instance_models]
    result_type = Type(type_model.id, type_model.name, base, attributes, instances, type_model.module)
    types_by_name[type_model.name] = result_type
    return result_type

def instance_model_to_instance(instance_model, types_by_name):
    return Instance(instance_model.id, types_by_name.get(instance_model.type.name), instance_model.name, instance_model.attributes)

def acl_model_to_acl(acl_model):
    return Acl(acl_model.type, acl_model.permission, acl_model.target_id, acl_model.org_id, acl_model.owner_id)

def item_model_to_item(org_id, item_model, types_by_name, children=True):
    type = types_by_name.get(item_model.type.name)
    if children:
        child_items = db.query(ItemModel).filter(ItemModel.parent_id == item_model.id, InstanceModel.org_id == org_id).all()
        return Item(item_model.owner_id,
                    item_model.org_id,
                    Instance(item_model.id, type, item_model.name),
                    item_model.id, 
                    item_model.name, 
                    item_model.attributes, 
                    [item_model_to_item(org_id, ci, types_by_name) for ci in child_items], 
                    [acl_model_to_acl(acl) for acl in item_model.acls])
    else:
        return Item(item_model.owner_id,
                    item_model.org_id,
                    Instance(item_model.id, type, item_model.name),
                    item_model.id, 
                    item_model.name, 
                    item_model.attributes, 
                    [], 
                    [acl_model_to_acl(acl) for acl in item_model.acls])

def get_type_attributes(item_type):
    attributes = item_type.attributes
    if item_type.base_id:
        base_attributes = get_type_attributes(item_type.base)
        if base_attributes and attributes:
            attributes = base_attributes | attributes
        elif base_attributes:
            attributes = base_attributes
    return attributes

def get_type_instances(type):
    instances = []
    if type.instances:
        instances.extend(type.instances)
    if type.base:
        instances.extend(get_type_instances(type.base))
    return instances

def build_item( item_id,
                instance,
                attributes,
                item_data,
                owner_id,
                org_id,
                parent_item_id=None):

    location = item_data.get('item_location')
    valid_from = item_data.get('item_valid_from')
    valid_to = item_data.get('item_valid_to')
    status = item_data.get('item_status')
    tags = item_data.get('item_tags')
    linked_item_id = item_data.get('item_linked_item_id')

    if not location:
        item = ItemModel( id=item_id,
                    name=instance.name,
                    attributes=attributes,
                    valid_from=valid_from,
                    valid_to=valid_to,
                    status=status,
                    tags=tags,
                    owner_id=owner_id,
                    org_id=org_id,
                    type_id=instance.type.id,
                    parent_id=parent_item_id,
                    linked_item_id=linked_item_id)
    else:
        item = ItemModel( id=item_id,
                    name=instance.name,
                    attributes=attributes,
                    valid_from=valid_from,
                    valid_to=valid_to,
                    status=status,
                    tags=tags,
                    owner_id=owner_id,
                    org_id=org_id,
                    location=location,
                    type_id=instance.type.id,
                    parent_id=parent_item_id,
                    linked_item_id=linked_item_id)
    
    db.add(item)
    db.commit()

    create_public_acl = instance.public

    if item_data and 'item_is_public' in item_data:
        if str(item_data['item_is_public']).lower() == 'true':
            create_public_acl = True

    if create_public_acl:
        acl = AclModel(id=str(uuid.uuid4()),type='public', org_id=org_id, item_id=item_id)
        db.add(acl)
        db.commit()
    
    # instances1 = instance.type.instances
    # instances2 = instance.type.base.instances
    instances = get_type_instances(instance.type)
    for child_instance in instances:
        attributes = get_type_attributes(child_instance.type)
        if attributes:
            if child_instance.attributes:
                attributes =  attributes | child_instance.attributes
        elif child_instance.attributes:
            attributes = child_instance.attributes
        
        child_item = build_item(  str(uuid.uuid4()),
                                  child_instance,
                                  attributes,
                                  {},
                                  owner_id,
                                  org_id,
                                  item.id)

    return item

def create_item_model(  db,
                        item_id,
                        item_type_name, 
                        item_name,
                        item_attributes,
                        item_location,
                        item_visibility,
                        item_tags,
                        item_is_public,
                        owner_id,
                        org_id,
                        parent_item_id=None,
                        item_valid_from=None,
                        item_valid_to=None,
                        item_status=None,
                        item_linked_item_id=None):

    item = None
    item_type = db.query(TypeModel).filter(TypeModel.name == item_type_name, TypeModel.org_id == org_id).first()
    if not item_type:
        item_type = TypeModel(id=str(uuid.uuid4()),
                         name=item_type_name,
                         owner_id=owner_id,
                         org_id=org_id)
        db.add(item_type)
        db.commit()

    if item_type:
        # attributes = item_attributes | item_type.attributes
        attributes = get_type_attributes(item_type)
        if attributes:
            if item_attributes:
                attributes =  attributes | item_attributes
        elif item_attributes:
            attributes = item_attributes 

        instance = create_instance_model(instance_id=item_id,
                            instance_name=item_name,
                            instance_icon=None,
                            instance_attributes=None,
                            instance_public=None,
                            owner_id=owner_id,
                            org_id=org_id,
                            instance_type_id=item_type.id)

        db.add(instance)
        db.commit()

        item_data = {"item_location": item_location, 
                     "item_visibility": item_visibility,
                     "item_tags": item_tags,
                     "item_is_public": item_is_public,
                     "item_valid_from": item_valid_from,
                     "item_valid_to": item_valid_to,
                     "item_status": item_status,
                     "item_linked_item_id": item_linked_item_id}
        
        item = build_item(item_id, instance, attributes, item_data, owner_id, org_id, parent_item_id)
    
    return item

def create_type_model(  db,
                        type_id,
                        type_name, 
                        type_icon,
                        type_attributes,
                        owner_id,
                        org_id,
                        type_module=None,
                        type_base_id=None):

    if type_base_id:
        type_base = db.query(TypeModel).filter(TypeModel.id == type_base_id, TypeModel.org_id == org_id).first()
        if not type_base:
            return None

    return TypeModel(id=type_id, name=type_name, icon=type_icon, attributes=type_attributes, module=type_module, base_id=type_base_id, owner_id=owner_id, org_id=org_id)


def create_instance_model(  instance_id,
                            instance_name,
                            instance_icon,
                            instance_attributes,
                            instance_public,
                            owner_id,
                            org_id,
                            instance_type_id,
                            instance_parent_type_id=None):
    return InstanceModel(id=instance_id,
                        name=instance_name,
                        icon=instance_icon,
                        attributes=instance_attributes,
                        public=instance_public,
                        owner_id=owner_id,
                        org_id=org_id,
                        type_id= instance_type_id,
                        parent_type_id=instance_parent_type_id)



from realnet.core.provider import OrgProvider, GroupProvider, AclProvider,AccountProvider, AppProvider
from realnet.core.type import Account, Authenticator, Org
from realnet.core.acl import AclType
from .utility import get_types_by_name, item_model_to_item, get_derived_types
from .models import Account as AccountModel, Org as OrgModel, Item as ItemModel, AccountGroup as AccountGroupModel, session as db

class SqlOrgProvider(OrgProvider, GroupProvider, AclProvider, AccountProvider, AppProvider):
    
    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_account_groups(self, account_id):
        return []

    def get_org_groups(self, org_id):
        return []

    # acl provider

    def is_item_public(self, item):
    
        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        return False

    def can_account_execute_item(self, account, item):
        
        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and acl.permission and  'e' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and  acl.permission and 'e' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and  acl.permission and 'e' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_message_item(self, account, item):

        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and  acl.permission and 'm' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  'm' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and acl.permission and  'm' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_read_item(self, account, item):

        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.user and acl.target_id == account.id and acl.permission and ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type == AclType.group and acl.target_id in group_ids and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_write_item(self, account, item):

        if [acl for acl in item.acls if
            acl.type == AclType.user and acl.target_id == account.id and acl.permission and 'w' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if
            acl.type == AclType.group and acl.target_id in group_ids and acl.permission and 'w' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type == AclType.org and acl.org_id == account.org_id and acl.permission and  'w' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True

    def can_account_delete_item(self, account, item):
        return item.owner_id == account.id and account.org_id == item.org_id

    # group provider

    def get_groups(self):
        pass

    # account provider

    def get_account(self):
        account = db.query(AccountModel).filter(AccountModel.id == self.account_id).first()
        if account:
            return Account(account.id, account.username, Org(account.org.id, account.org.name))
        return None

    def get_org(self):
        org = db.query(OrgModel).filter(OrgModel.id == self.org_id).first()
        if org:
            return Org(org.id, org.name)
        return None

    # app provider

    def get_apps(self, module):
        account = db.query(AccountModel).filter(AccountModel.id == self.account_id).first()
        if account:
            tbn = get_types_by_name(self.org_id)
            application_type = tbn.get('App')
            application_type_ids = set(get_derived_types(self.org_id, [application_type.id]) + [application_type.id])
            role_apps = [app.app for ar in account.roles for app in ar.role.apps ]
            owned_apps = db.query(ItemModel).filter(ItemModel.owner_id == self.account_id, ItemModel.type_id.in_(application_type_ids) ).all()
            app_ids = {app.id:app for app in role_apps}
            for owned_app in owned_apps:
                if not owned_app.id in app_ids:
                    role_apps.append(owned_app)
                    app_ids[owned_app.id] = owned_app

            return [item_model_to_item(self.org_id, app, tbn) for app in role_apps ]
        return []






import os
from realnet.shell import ProtoShell, ProtoCmd
from realnet.core.provider import ContextProvider, Context
from realnet.core.hierarchy import import_structure_from_resource

from realnet.provider.generic.endpoint import GenericEndpointProvider
from realnet.provider.generic.resource import GenericResourceProvider
from realnet.provider.generic.importer import GenericImportProvider
from realnet.provider.sql.type import SqlTypeProvider
from realnet.provider.sql.postgres.item import PostgresItemProvider
from realnet.provider.aws.data import S3DataProvider
from realnet.provider.sql.org import SqlOrgProvider
from realnet.provider.sql.orgs import SqlOrgsProvider
from realnet.provider.sql.init import SqlInitProvider
from realnet.provider.sql.client import SqlClientProvider

from realnet.runner.http.runner import HttpRunner

class StandardContextProvider(ContextProvider):
    
    def context(self, org_id, account_id):
        org_provider = SqlOrgProvider(org_id, account_id)
        return Context( SqlTypeProvider(org_id, account_id),
                        PostgresItemProvider(org_id, account_id),
                        S3DataProvider(),
                        org_provider,
                        org_provider,
                        org_provider,
                        None,
                        GenericEndpointProvider(),
                        None,
                        org_provider,
                        org_provider,
                        GenericResourceProvider(),
                        GenericImportProvider(),
                        SqlClientProvider(org_id, account_id))

        
    def get_orgs(self):
        return SqlOrgsProvider().get_orgs()

    def get_org_by_id(self,id):
        return SqlOrgsProvider().get_org_by_id(id)

    def get_org_by_name(self,name):
        return SqlOrgsProvider().get_org_by_name(name)

    def get_org_authenticators(self, org_id):
        return SqlOrgsProvider().get_org_authenticators(org_id)

    def get_account_by_id(self,id):
        return SqlOrgsProvider().get_account_by_id(id)

    def check_password(self, org_id, account_id, password):
        return SqlOrgsProvider().check_password(org_id, account_id, password)

    def get_org_clients(self, org_id):
        return SqlOrgsProvider().get_org_clients(org_id)

    def get_org_client(self, org_id, client_id):
        return SqlOrgsProvider().get_org_client(org_id, client_id)

    def get_public_apps(self, org_id):
        return SqlOrgsProvider().get_public_apps(org_id)

    def get_public_types(self, org_id):
        return SqlOrgsProvider().get_public_types(org_id)

    def get_public_forms(self, org_id):
        return SqlOrgsProvider().get_public_forms(org_id)

    def get_public_item(self, id):
        return SqlOrgsProvider().get_public_item(id)

    def get_public_orgs(self):
        return SqlOrgsProvider().get_public_orgs()

    def initialize(self, org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri):
        return SqlInitProvider().initialize(org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri)

class Start(ProtoCmd):
    
    def __init__(self):
        super().__init__('start',
                         'start realnet server')

    def add_arguments(self, parser):
        # parser.add_argument('--item', help='specify the name of the item to be served, default is root home folder')
        pass

    def run(self, args):
        contextProvider = StandardContextProvider()
        runner = HttpRunner()
        runner.run(contextProvider)
        # cfg = Config()
        # if args.item:
        #     app.config['ROOT_ITEM'] = args.item
        # app.run(cfg.get_server_host(), cfg.get_server_port())


class Initialize(ProtoCmd):
    
    def __init__(self):
        super().__init__('initialize',
                         'initialize api server')

    def add_arguments(self, parser):
        parser.add_argument('--name', help='specify the tenant name', default=os.getenv('REALNET_NAME'))
        parser.add_argument('--username', help='specify the root username', default=os.getenv('REALNET_USERNAME'))
        parser.add_argument('--password', help='specify the root password', default=os.getenv('REALNET_PASSWORD'))
        parser.add_argument('--email', help='specify the root email', default=os.getenv('REALNET_EMAIL'))
        parser.add_argument('--uri', help='specify the tenant uri', default=os.getenv('REALNET_URI'))
        parser.add_argument('--redirect_uri', help='specify the tenant redirect uri', default=os.getenv('REALNET_REDIRECT_URI'))
        parser.add_argument('--mobile_redirect_uri', help='specify the tenant mobile redirect uri', default=os.getenv('REALNET_MOBILE_REDIRECT_URI'))

    def run(self, args):
        context_provider = StandardContextProvider()
        account = context_provider.initialize(args.name, args.username, args.email, args.password, args.uri, args.redirect_uri, args.mobile_redirect_uri)
        if account:
            context = context_provider.context(account.org.id, account.id)
            if context:
                import_structure_from_resource(context, 'static/initialization/core.json')
        

        # with app.app_context():
        #    db.create_all()
        #    response = initialize_server(args.name, args.type, args.username, args.email, args.password, args.uri, args.redirect_uri, args.mobile_redirect_uri)
        #    print(response)

class Server(ProtoShell):
    
    def __init__(self):
        super().__init__('server', [
            Start(), Initialize()
        ], 'realnet server')

    def add_arguments(self, parser):
        pass

    def get_help(self):
        return 'realnet server help'


from realnet.core.provider import ClientProvider
from realnet.core.type import Client, Org
from .models import Client as ClientModel, session as db
from sqlalchemy import or_

class SqlClientProvider(ClientProvider):

    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_clients(self, module):
        return [Client( client.id, 
                        client.name, 
                        Org(client.org.id, client.org.name), 
                        client.attributes) for client in db.query(ClientModel).filter(ClientModel.org_id == self.org_id).all()]

    def get_client(self, client_id):
        return db.query(ClientModel).filter(or_(ClientModel.client_id == client_id, ClientModel.name == client_id),ClientModel.org_id == self.org_id).first()


from realnet.shell import ProtoShell

class Get(ProtoShell):
    
    def __init__(self):
        super().__init__('get', [], 'realnet get')

    def add_arguments(self, parser):
        pass

    def run(self, args):
        print('this is get')


from realnet.shell import ProtoShell

class Create(ProtoShell):
    
    def __init__(self):
        super().__init__('create', [], 'realnet create')

    def add_arguments(self, parser):
        pass

    def run(self, args):
        print('this is create')


from realnet.core.provider import EndpointProvider
from realnet.core.type import Endpoint


class GenericEndpointProvider(EndpointProvider):

    def __init__(self):
        pass

    def get_endpoints(self, module):
        account = module.get_account()
        return [Endpoint(e) for e in module.find_items({'types': ['Endpoint'], 'any_level': 'true'}) if module.can_account_read_item(account, e)]

    def get_endpoint(self, module, endpoint_name):
        account = module.get_account()
        endpoints = [Endpoint(e) for e in module.find_items({'types': ['Endpoint'], 'keys': ['path'], 'values': [endpoint_name], 'any_level': 'true'}) if module.can_account_read_item(account, e)]
        return next(iter(endpoints), None)


from realnet.core.provider import RolesProvider
from realnet.core.type import Role, Org, App
from .models import Role as RoleModel, session as db

class SqlRolesProvider(RolesProvider):

    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_roles(self, module):
        return [Role(   role.id, 
                        role.name, 
                        Org(role.org.id, role.org.name), 
                        [App(app.app.id, app.app.name) for app in role.apps ]) for role in db.query(RoleModel).filter(RoleModel.org_id == self.org_id).all()]

    def create_role(self, **kwargs):
        pass

    def delete_role(self, id):
        pass

    def update_role(self, id, **kwargs):
        pass


from realnet.core.config import Config
from realnet.core.provider import DataProvider
from realnet.core.type import Data

import boto3

cfg = Config()

class S3DataProvider(DataProvider):
    
    def __init__(self, bucket_name = cfg.get_s3_bucket()):
        self.bucket_name = bucket_name
        self.session = boto3.Session(
            aws_access_key_id=cfg.get_s3_key(),
            aws_secret_access_key=cfg.get_s3_secret(),
            region_name=cfg.get_s3_region()
        )

    def get_data(self, id):
        s3 = self.session.resource('s3')
        bucket = s3.Bucket(self.bucket_name)
        try:
            s3_obj = bucket.Object(id)
            if s3_obj:
                s3_obj_data = s3_obj.get()
                return Data(id, s3_obj_data['ContentType'], s3_obj_data['ContentLength'], s3_obj_data['Body'].read())
            else:
                return None
        except:
            return None

    def update_data(self, id, storage):
        s3 = self.session.resource('s3')
        bucket = s3.Bucket(self.bucket_name)
        bucket.Object(id).put(Body=storage)

    def delete_data(self, id):
        s3 = self.session.resource('s3')
        bucket = s3.Bucket(self.bucket_name)
        bucket.Object(id).delete()

    def get_data_upload_url(self, id):
        s3 = self.session.client('s3')
        try:
            response = s3.generate_presigned_post(self.bucket_name,
                                                  id,
                                                  ExpiresIn=3600)
            return response
        except Exception as e:
            print(e)
            return None

    def confirm_data_upload(self, id):
        pass


from flask import Flask, Response, redirect, render_template, render_template_string, request, jsonify, Blueprint, send_file, session
from flask_cors import CORS
from flask_bootstrap import Bootstrap

import jinja2

import logging

from realnet.core.config import Config
from .auth import config_oauth
from .router import router_bp

def create_app(contextProvider):
    logging.basicConfig(level=logging.DEBUG)

    cfg = Config()

    app = Flask(__name__)
    CORS(app)
    # import realnet_server.wsgi

    app.secret_key = cfg.get_app_secret()
    app.config['BOOTSTRAP_SERVE_LOCAL'] = True
    app.config['REALNET_CONTEXT_PROVIDER'] = contextProvider


    app.jinja_loader = jinja2.ChoiceLoader([
        app.jinja_loader,
        jinja2.PackageLoader('realnet') # in the same folder will search the 'templates' folder
    ])
    
    bootstrap = Bootstrap(app)

    config_oauth(app)

    app.register_blueprint(router_bp)

    return app


from abc import ABC, abstractmethod
from .acl import Acl, AclType
from .provider import *

class Resource(ABC):
    
    @abstractmethod
    def get(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def post(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def put(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def delete(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def message(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def run(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def get_data(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def update_data(self, module, args, path=None, content_type='text/html'):
        pass

    @abstractmethod
    def delete_data(self, module, args, path=None, content_type='text/html'):
        pass
    
class Type:
    
    def __init__(self, id, name, base=None, attributes=dict(), instances=[], module=None, types=dict()):
        self.id = id
        self.name = name
        self.base = base
        self._attributes = attributes
        self._instances = instances
        self.module = module
        self.types = types

    def _get_attributes(self):
        if self.base:
            return  self.base.attributes | self._attributes
        else:
            return self._attributes

    def _del_attributes(self):
        self._attributes = dict()

    def _set_attributes(self, value):
        self._attributes = value

    attributes = property(
        fget = _get_attributes,
        fset = _set_attributes,
        fdel = _del_attributes
    )

    def _get_instances(self):
        if self.base:
            return self.base.instances + self._instances
        else:
            return self._instances

    def _del_instances(self):
        self._instances = []

    def _set_instances(self, value):
        self._instances = value

    instances = property(
        fget = _get_instances,
        fset = _set_instances,
        fdel = _del_instances
    )

    def is_derived_from(self, type_name):
        if type_name.lower() == self.name.lower():
            return True
        elif self.base:
            return self.base.is_derived_from(type_name)
        return False

    def to_dict(self):
        base_data = None
        if self.base:
            base_data = self.base.to_dict()
        instances = []
        internal_instances = self.instances
        if internal_instances:
            instances = [i.to_dict() for i in internal_instances]

        return {
            'id': self.id,
            'name': self.name,
            'base': base_data,
            'attributes': self.attributes,
            'instances': instances
        }


class Instance(Type):
    
    def __init__(self, id, type, name, attributes=dict()):
        self.id = id
        self.type = type
        self.name = name
        self._attributes = attributes

    def _get_attributes(self):
        return  self.type.attributes | self._attributes

    def _del_attributes(self):
        self._attributes = dict()

    def _set_attributes(self, value):
        self._attributes = value

    attributes = property(
        fget = _get_attributes,
        fset = _set_attributes,
        fdel = _del_attributes
    )

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'type': self.type.name,
            'attributes': self.attributes
        }



class Item(Instance):
    
    def __init__(self, owner_id, org_id, instance, id, name, attributes=dict(), items=[], acls=[]):
        self.owner_id = owner_id
        self.org_id = org_id
        self.id = id
        self.name = name
        self.instance = instance
        self._attributes = attributes
        self.items = items
        self.acls = acls

    def _get_attributes(self):
        return  self.instance.attributes | self._attributes if self._attributes else self.instance.attributes

    def _del_attributes(self):
        self._attributes = dict()

    def _set_attributes(self, value):
        self._attributes = value

    attributes = property(
        fget = _get_attributes,
        fset = _set_attributes,
        fdel = _del_attributes
    )

    def items_of_type(self, type_name):
        return [i for i in self.items if i.instance.type.is_derived_from(type_name)]

    def is_of_type(self, type_name):
        return self.instance.type.is_derived_from(type_name)

    def get_base_types(self):
        pass

    def get_child_instances(self):
        pass

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'type': self.instance.type.to_dict(),
            'attributes': self.attributes,
            'items': [i.to_dict() for i in self.items]
        }


class Data:

    def __init__(self, id, mimetype, length, bytes):
        self.id = id
        self.mimetype = mimetype
        self.length = length
        self.bytes = bytes

class Authenticator:
    
    def __init__(self, name, url):
        self.name = name
        self.url = url

class Func:
    
    def __init__(self, item=None, callback = None):
        self.item = item
        self.callback = callback

    def invoke(self, module, args, path=None, content_type='text/html'):
        if self.callback:
            return self.callback(module, args, path, content_type)
        elif self.item:
            code = self.item.attributes.get('code')
            if not code:
                data = module.get_data(self.item.id)
                if data:
                    code = data.decode('utf-8')
                    result = dict()
                    exec(code, {'module': module, 'args': args, 'path': path, 'content_type': content_type, 'item': self.item, 'result': result})
                    return result['response']
        else:
            return None


class Endpoint:
    
    def __init__(self, item):
        self.item = item

    def invoke(self, module, method, args, path=None, content_type='text/html'):
        
        func = module.get_resource_method(module, self.item.attributes['resource'], method.lower())
        
        if func:
            return func.invoke(module, args, path, content_type)

        return None

class Task:

    def __init__(self):
        pass


class List:
    
    def __init__(self):
        pass

class Error:
    
    def __init__(self):
        pass

class Group:
    
    def __init__(self):
        pass

class Org:
    
    def __init__(self, id, name):
        self.id = id
        self.name = name

class Account:
    
    def __init__(self, id, name, org):
        self.id = id
        self.name = name
        self.org = org

    def get_user_id(self):
        return self.id

class App:
    
    def __init__(self, id, name):
        self.id = id
        self.name = name

class Role:
    
    def __init__(self, id, name, org, apps):
        self.id = id
        self.name = name
        self.org = org
        self.apps = apps

class Client:
    
    def __init__(self, id, name, org, attributes):
        self.id = id
        self.name = name
        self.org = org
        self.apps = attributes








from .shell import Shell
from .cmd import Cmd


class ProtoShell(Shell):

        def __init__(self, name, commands, helpmsg):
            super().__init__(name)
            self.commands = commands
            self.helpmsg = helpmsg

        def get_commands(self):
            return self.commands

        def add_arguments(self, parser):
            pass

        def get_help(self):
            return self.helpmsg


class ProtoCmd(Cmd):

    def __init__(self,
                 name,
                 helpmsg,
                 lambda_of_args=lambda args: args):
        super().__init__(name)
        self.lambda_of_args = lambda_of_args
        self.helpmsg = helpmsg

    def add_arguments(self, parser):
        pass

    def get_help(self):
        return self.helpmsg

    def run(self, args):
        return self.lambda_of_args(args)


from realnet.core.provider import InitializationProvider
from realnet.core.type import Org, Account
from realnet.provider.sql.models import initialize

class SqlInitProvider(InitializationProvider):

    def initialize(self, org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri):
        accouunt_model = initialize(org_name, admin_username, admin_email, admin_password, uri, redirect_uri, mobile_redirect_uri)
        return Account(accouunt_model.id, accouunt_model.username, Org(accouunt_model.org.id, accouunt_model.org.name))


import json
import os
import sys
import uuid

from .type import Type

def traverse_instance(instances, instance, parent_type_name):
        for inst in instance.get('instances', []):
            instances.append({ "instance": inst, "parent_type_name": parent_type_name})
            traverse_instance(instances, inst, inst.get('type'))

def collect_inheritance_hierarchy(children, td, instances):
    name = td.get('name')
    derived_types = children.get(name)
    if derived_types:
        for derived_type in derived_types:
            for instance in td.get('instances', []):
                instances.append({ "instance": instance, "parent_type_name": derived_type['name']})
                # logging.getLogger().info('*** {} is derived from {} ***'.format(derived_type['name'], name))
            collect_inheritance_hierarchy(children, derived_type, instances)



def collect_type_dependencies(module, td):
    name = td.get('name')
    deps = set()
    base_name = td.get('base')
    if base_name:
        deps.add(base_name)
    for instance in td.get('instances', []):
        deps.add(instance['type'])
    return deps

def collect_type_defs(module, type_data, type_defs):
    for td in type_data:
        name = td.get('name')
        deps = collect_type_dependencies(module, td)
        if not name in type_defs: 
            type_defs[name] = {'type': td, 'deps': deps}
            child_types = td.get('types')
            if child_types:
                collect_type_defs(module, child_types, type_defs)

def build_type(module, type, existing_types_by_name):
    base_id = None
    base_name = type.get('base')
    if base_name:
        base_type = existing_types_by_name.get(base_name)
        if base_type:
            base_id = base_type.id

    params = type | {'base_id': base_id}

    created_type = module.create_type(**params)

    for instance in type.get('instances', []):
        attributes = instance.get('attributes', dict())
        is_public = instance.get('public')
        if is_public:
            is_public = is_public.lower() in ['true', 'True', '1']
        type_id = existing_types_by_name[instance['type']].id
        if type_id:
            module.create_instance(**{  'id': str(uuid.uuid4()),
                                        'name': instance['name'],
                                        'attributes': attributes,
                                        'public': is_public,
                                        'type_id': type_id,
                                        'parent_type_id': created_type.id})

    existing_types_by_name[created_type.name] = created_type

def build_type_tree(module, queue, built, existing_types_by_name):
    to_build = []
    on_hold = []

    for element in queue:
        if element['deps'] <= built:
            to_build.append(element)
        else:
            on_hold.append(element)

    for element in to_build:
        build_type(module, element['type'], existing_types_by_name)
        built.add(element['type']['name'])

    if on_hold:
        build_type_tree(module, on_hold, built, existing_types_by_name)




def import_types(module, type_data):
    type_defs = dict()
    collect_type_defs(module, type_data, type_defs)
    existing_types_by_name = {t.name:t for t in module.get_types()}

    build_type_tree(module, type_defs.values(), set(existing_types_by_name.keys()), existing_types_by_name)



def traverse_item(module, item_ids, items_by_id, children_by_id, item):
    parent_id = item.get('parent_id')
    if parent_id:
        module.create_item(item | {'parent_id': parent_id})
    else:
        module.create_item(item)
    for child in children_by_id.get(item['id'], []):
        traverse_item(module, item_ids, items_by_id, children_by_id, items_by_id.get(child))

def traverse_instance(instances, instance, parent_type_name):
    for inst in instance.get('instances', []):
        instances.append({ "instance": inst, "parent_type_name": parent_type_name})
        traverse_instance(instances, inst, inst.get('type'))


def get_base_types(module, type, base_types):
    base_types.add(type)
    if type and type.base:
        return get_base_types(module, type.base, base_types)
    return base_types

def get_instance_types(module, instance):
    return get_base_types(module, instance.type, set())


def import_items(module, items):

    for item in items:
        item_attributes = item.get('attributes', dict())
        item_parent_id = item.get('parent_id')
        item_id = item.get('id', str(uuid.uuid4()))
        item_location = item.get('location')
        if item_location and not isinstance(item_location,str):
            item_location = json.dumps(item_location)
        item_type = item.get('type')
        item_name = item.get('name')
        item_is_public = str(item.get('public')).lower() == 'true'
        # item_visibility = VisibilityType.visible
        item_tags = item.get('tags')
        
        item_data = dict()

        if item_parent_id == item_id:
            item_data = {"id": item_id,
                        "type": item_type,
                        "attributes": item_attributes,
                        "name": item_name,
                        "location": item_location,
                        #"visibility": item_visibility,
                        "tags": item_tags,
                        "public": item_is_public}
        else:
            item_data = {"id": item_id,
                        "type": item_type,
                        "attributes": item_attributes,
                        "name": item_name,
                        "location": item_location,
                        # "visibility": item_visibility,
                        "tags": item_tags,
                        "parent_id": item_parent_id,
                        "public": item_is_public}
        created_item = module.create_item(**item_data)

def import_structure_from_resource(context, path):
    with open(os.path.join(os.path.abspath(os.path.join(os.path.dirname(sys.modules[__name__].__file__), os.pardir)),
                           path), 'r') as f:
        data = json.load(f)
        if data:
            type_data = data.get('types')
            if type_data:
                import_types(context, type_data)

            item_data = data.get('items')
            if item_data:
                import_items(context, item_data)


import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="realnet",
    version="0.0.42",
    description="Realnet command line interface",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/virtual-space/realnet",
    author="Marko Laban",
    author_email="marko.laban@l33tsystems.com",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    package_data={  'realnet.static.initialization': ['*'],
                    'realnet.templates': ['*'],
                    'realnet.runner.http.static': ['scripts/ace/*', 'font-awesome/css/*', 'font-awesome/fonts/*']},
    install_requires=["python-dotenv",
                      "boto3",
                      "Flask==2.2.2",
                      "SQLAlchemy==1.4.23",
                      "psycopg2-binary",
                      "sqlalchemy-serializer",
                      "shapely==1.8.5",
                      "GeoAlchemy2",
                      "Werkzeug",
                      "authlib",
                      "flask-cors",
                      "bootstrap-flask",
                      "requests"],
    entry_points={
        "console_scripts": [
            "realnet=realnet.__main__:main",
        ]
    },
)



from realnet.shell import ProtoShell

class Runner(ProtoShell):
    
    def __init__(self):
        super().__init__('runner', [], 'realnet runner')

    def add_arguments(self, parser):
        pass

    def run(self, args):
        print('this is runner')


from abc import abstractmethod
from .client import Client
import argparse


class Cmd(Client):

    def __init__(self, name):
        self.name = name

    def get_client(self):
        return self

    @abstractmethod
    def add_arguments(self, parser):
        pass

    @abstractmethod
    def get_help(self):
        return None

    def setup(self, subparsers, parent=None):
        parser = subparsers.add_parser(self.name, help=self.get_help())
        self.add_arguments(parser)
        return parser

    def repl(self):
        ui = input('[{0}]'.format(self.name))
        self.run(ui)
        self.repl()

    def __call__(self, args):
        parser = argparse.ArgumentParser(prog=self.name)
        self.add_arguments(parser)
        return self.run(parser.parse_args(args))


import json

from typing import cast
from urllib.parse import unquote
import uuid

from sqlalchemy.sql import func, or_
from sqlalchemy.dialects.postgresql import ARRAY

from realnet.core.provider import ItemProvider
from realnet.core.acl import Acl, AclType
from ..utility import get_types_by_name, item_model_to_item, create_item_model, get_derived_types
from ..models import session as db, Item as ItemModel, Type as TypeModel, VisibilityType


class PostgresItemProvider(ItemProvider):

    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_items(self):
        pass

    def get_item(self, id, children=False):
        item_model = db.query(ItemModel).filter(ItemModel.id == id, ItemModel.org_id == self.org_id).first()
        tbn = get_types_by_name(self.org_id)
        return item_model_to_item(self.org_id, item_model, tbn, children)

    def delete_item(self, id):
        try:
            if id:
                target_item = db.query(ItemModel).filter(ItemModel.id == id, ItemModel.org_id == self.org_id).first()
                if target_item:
                    db.delete(target_item)
                    db.commit()
        except Exception as e:
            print(e)

    def update_item(self, id, **kwargs):
        item_model = db.query(ItemModel).filter(ItemModel.id == id, ItemModel.org_id == self.org_id).first()
        if item_model:
            for key, value in kwargs.items():
                if key == 'name':
                    item_model.name = value
                elif key == 'attributes':
                    item_model.attributes = value
            db.commit()

    def create_item(self, **kwargs):
        item_id=str(uuid.uuid4())
        item_name = None
        item_type_name = None
        item_attributes = None
        item_parent_id = None
        item_location = None
        item_visibility = None
        item_tags = None
        item_is_public = False
        item_valid_from = None
        item_valid_to = None
        item_status = None
        item_linked_item_id = None

        for key, value in kwargs.items():
            # print("%s == %s" % (key, value))
            if key == 'name':
                item_name = value
            elif key == 'id':
                item_id = value
            elif key == 'type_id':
                item_type_id = value
            elif key == 'type':
                type = db.query(TypeModel).filter(TypeModel.name == value, TypeModel.org_id == self.org_id).first()
                if type:
                    item_type_name = type.name
            elif key == 'attributes':
                item_attributes = value
            elif key == 'tags':
                item_tags = value
            elif key == 'location' and value:
                data = json.loads(value)
                if isinstance(data,str):
                    data = json.loads(data)
                if data['type'] == 'Point':
                    item_location = 'SRID=4326;POINT({0} {1})'.format(data['coordinates'][0], data['coordinates'][1])
                else:
                    item_location = 'SRID=4326;POLYGON(('
                    for ii in data['coordinates'][0]:
                        item_location = item_location + '{0} {1},'.format(ii[0], ii[1])
                    item_location = item_location[0:-1] + '))'
            elif key == 'visibility':
                item_visibility = value
            elif key == 'public':
                item_is_public = value == True
            elif key == 'valid_from':
                item_valid_from = value
            elif key == 'valid_to':
                item_valid_to = value
            elif key == 'status':
                item_status = value
            elif key == 'linked_item_id':
                item_linked_item_id = value
            elif key == 'parent_id':
                item_parent_id = value

        item = create_item_model(
            db=db,
            item_id=item_id,
            item_type_name=item_type_name,
            item_name=item_name,
            item_tags=item_tags,
            owner_id=self.account_id,
            org_id=self.org_id,
            parent_item_id=item_parent_id,
            item_attributes=item_attributes,
            item_visibility=item_visibility,
            item_location=item_location,
            item_is_public=item_is_public,
            item_valid_from=item_valid_from,
            item_valid_to=item_valid_to,
            item_status=item_status,
            item_linked_item_id=item_linked_item_id)

        db.add(item)
        db.commit()

        tbn = get_types_by_name(self.org_id)

        return item_model_to_item(self.org_id, item, tbn)

    def find_items(self, data):

        op_or = False

        type_names = data.get('type_names')
        if type_names:
            data['type_names'] = type_names
        else:
            types = data.get('types')
            if types:
                data['type_names'] = types
                type_names = types
        

        public = data.get('public')
        if public:
            data['public'] = public
        
        tags = data.get('tags')
        if tags:
            data['tags'] = tags

        status = data.get('status')
        if status:
            data['status'] = status

        name = data.get('name')
        if name:
            data['name'] = name
        
        parent_id = data.get('parent_id')
        if parent_id:
            data['parent_id'] = parent_id
        else:
            if 'any_level' in data and str(data['any_level']).lower() == 'true': 
                pass
            else:
                data['parent_id'] = None

        location = data.get('location')
        if location:
            data['location'] = location

        valid_from = data.get('valid_from')
        if valid_from:
            data['valid_from'] = valid_from

        valid_to = data.get('valid_to')
        if valid_to:
            data['valid_to'] = valid_to

        status = data.get('status')
        if status:
            data['status'] = status

        # TODO below

        keys = data.get('keys')
        if keys:
            data['keys'] = keys

        values = data.get('values')
        if values:
            data['values'] = values

        visibility = data.get('visibility')
        if visibility:
            data['visibility'] = visibility

        op = data.get('op')

        if op:
            if op.lower() == 'or':
                op_or = True

        conditions = []

        if type_names:
            type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name.in_(type_names), TypeModel.org_id == self.org_id).all()]
            derived_type_ids = get_derived_types(self.org_id, type_ids)
            conditions.append(ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))))

        if tags:
            if isinstance(tags, list):
                conditions.append(ItemModel.tags.contains(cast(ARRAY(db.String), tags)))
            else:
                conditions.append(ItemModel.tags.contains([tags]))
        
        if name:
            conditions.append(ItemModel.name.ilike('{}%'.format(unquote(str(name)))))
        
        if parent_id:
            conditions.append(ItemModel.parent_id == parent_id)
        else:
            if 'any_level' in data and str(data['any_level']).lower() == 'true': 
                pass
            else:
                conditions.append(ItemModel.parent_id == None)

        if location:
            loc_data = json.loads(location)
            if isinstance(loc_data,str):
                    loc_data = json.loads(loc_data)
            if loc_data['type'] == 'Point':
                item_location = 'SRID=4326;POINT({0} {1})'.format(loc_data['coordinates'][0], loc_data['coordinates'][1])
                range = (0.00001) * float(500) #converting from meters to lng/lat scale
                conditions.append(func.ST_DWithin(ItemModel.location, item_location, range))
            else:
                item_location = 'SRID=4326;POLYGON(('
                for ii in loc_data['coordinates'][0]:
                    item_location = item_location + '{0} {1},'.format(ii[0], ii[1])
                item_location = item_location[0:-1] + '))'
                conditions.append(func.ST_Within(ItemModel.location, item_location))

        if valid_from:
            conditions.append(ItemModel.valid_from >= valid_from)

        if valid_to:
            conditions.append(ItemModel.valid_to <= valid_to)

        if status:
            conditions.append(ItemModel.status == status)

        
        if keys and values:
            if len(keys) > 1 and op_or:
                subconditions = []
                for kv in zip(keys, values):
                    subconditions.append(ItemModel.attributes.op('->>')(kv[0]) == kv[1])
                
                conditions.append(or_(*subconditions))
            else:
                for kv in zip(keys, values):
                    # conditions.append(ItemModel.attributes[kv[0]].astext == kv[1])
                    conditions.append(ItemModel.attributes.op('->>')(kv[0]) == kv[1])

        if visibility:
            conditions.append(ItemModel.visibility == VisibilityType[visibility])

        result_item_models = []

        if public:
            conditions.append(ItemModel.acls.any(Acl.type == AclType.public))
            result_item_models = ItemModel.query.filter(*conditions).all()
        else:
            if not conditions:
                return []
            else:
                result_item_models = db.query(ItemModel).filter(*conditions).all()

        tbn = get_types_by_name(self.org_id)
        items = [item_model_to_item(self.org_id, i, tbn) for i in result_item_models]
        return items



from flask import render_template
from realnet.resource.items.items import Items

class Public(Items):
    
    def get_endpoint_name(self):
        return 'public'

    def get(self, module, args, path=None, content_type='text/html'):
        if path:
            path_segments = path.split('/')

        if path_segments:
            subpath = path_segments[0]
            print(subpath)
        
        return self.render_item(module, args, path, content_type)
    